/*
 * Copyright (c) 1996-2002 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

// $Id: pdwinnt.C,v 1.42 2002/01/08 19:59:30 pcroth Exp $
#include <iomanip.h>
#include "dyninstAPI/src/symtab.h"
#include "common/h/headers.h"
#include "dyninstAPI/src/os.h"
#include "dyninstAPI/src/process.h"
#include "dyninstAPI/src/pdThread.h"
#include "dyninstAPI/src/stats.h"
#include "common/h/Types.h"
#include "dyninstAPI/src/showerror.h"
#include "dyninstAPI/src/instPoint.h"

#ifndef BPATCH_LIBRARY
#include "paradynd/src/main.h"
#include "paradynd/src/init.h"
#endif

#ifdef BPATCH_LIBRARY
/* XXX This is only needed for emulating signals. */
#include "BPatch_thread.h"
#include "nt_signal_emul.h"
#endif


//ccw  27 july 2000 : dummy methods to get the thing to compile before i add
//the remoteDevice : 29 mar 2001
#ifdef mips_unknown_ce2_11
#include <stdio.h> //for wprintf

void kludgeWCHAR(const char *str8, WCHAR *str16){
	int k;

	k=0;
	while(str8[k]!='\0'){
		str16[k] = str8[k];
		k++;
	}
	str16[k] = '\0';
	str16[k+1]='\0';

}

//int GetThreadContext(HANDLE hThread, w32CONTEXT *lpContext){
//	return 0;
//}

//int SetThreadContext(HANDLE hThread, const w32CONTEXT *lpContext){
//	return 0;
//}
#endif

extern bool isValidAddress(process *proc, Address where);
extern process *findProcess(int);

//HANDLE kludgeProcHandle;

void printSysError(unsigned errNo) {
    char buf[1000];
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, errNo, 
		  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		  buf, 1000, NULL);
    fprintf(stderr, "*** System error [%d]: %s\n", errNo, buf);
    fflush(stderr);
}


// check if a file handle is for kernel32.dll
static bool kludge_isKernel32Dll(HANDLE fileHandle, string &kernel32Name) {
    static DWORD IndxHigh, IndxLow;
    static bool firstTime = true;
    BY_HANDLE_FILE_INFORMATION info;
    static string kernel32Name_;

    if (firstTime) {
	HANDLE kernel32H;
	firstTime = false;
	char sysRootDir[MAX_PATH+1];
	if (GetSystemDirectory(sysRootDir, MAX_PATH) == 0)
	  assert(0);
	kernel32Name_ = string(sysRootDir) + "\\kernel32.dll";
	kernel32H = CreateFile(kernel32Name_.string_of(), GENERIC_READ, 
			       FILE_SHARE_READ, NULL, OPEN_EXISTING, NULL, NULL);
	assert(kernel32H);
	if (!GetFileInformationByHandle(kernel32H, &info)) {
	    printSysError(GetLastError());
	    assert(0);
	}
	IndxHigh = info.nFileIndexHigh;
	IndxLow = info.nFileIndexLow;
	CloseHandle(kernel32H);
    }

    if (!GetFileInformationByHandle(fileHandle, &info))
	return false;

    if (info.nFileIndexHigh==IndxHigh && info.nFileIndexLow==IndxLow) {
      kernel32Name = kernel32Name_;
      return true;
    }
    return false;
}


void dumpMem(process *p, void * addr, unsigned nbytes) {
    unsigned char *buf = new unsigned char[nbytes];
    memset(buf, 0, nbytes);
    function_base *f;
    assert(buf);

    if (f = p->findFuncByAddr((Address)addr))
    {
        printf("Function %s, addr=0x%lx, sz=%d\n", 
                f->prettyName().string_of(),
                f->getAddress(p),
                f->size());
    }
    p->readDataSpace((void *)((unsigned)addr-32), nbytes, buf, true);
    printf("## 0x%lx:\n", (unsigned)addr-32);
    for (unsigned u = 0; u < nbytes; u++)
    {
	    printf(" %x", buf[u]);
    }
    printf( "\n" );
    p->readDataSpace(addr, nbytes, buf, true);
    printf("## 0x%lx:\n", addr);
    for (unsigned u1 = 0; u1 < nbytes; u1++)
    {
	    printf(" %x", buf[u1]);
    }
    printf( "\n" );
}



//
// walkStackFrame
//
// Try to walk one more frame in the stack of the indicated process.
// Assumes that the STACKFRAME argument is either the result
// from a previous StackWalk call, or has been initialized 
// to the context from a GetThreadContext call.
//
inline
BOOL
walkStackFrame( HANDLE hProc, HANDLE hThread, STACKFRAME* psf )
{
	return StackWalk( IMAGE_FILE_MACHINE_I386,
                        hProc,
                        hThread,
		                psf,
                        NULL,
		                NULL,
                        SymFunctionTableAccess,
                        SymGetModuleBase,
		                NULL);
}


//
// findFunctionFromAddress
// 
// Finds the function corresponding to the given address.
// Handles situations where the address is within instrumentation,
// and situations where the function has been relocated.
//
function_base*
findFunctionFromAddress( process* proc, Address addr )
{
    function_base* fp = NULL;

    instPoint* ip = findInstPointFromAddress( proc, addr );
    if( ip != NULL )
    {
        // the address is within a tramp
        // find the function in which the tramp was installed
        fp = (function_base*)( ip->iPgetFunction() );
    }
    else
    {
        // the address isn't within a tramp
        // check if it is a known function (relocated or otherwise)
        fp = (function_base*)( proc->findFuncByAddr( addr ) );
    }
    return fp;
}



#ifdef i386_unknown_nt4_0 //ccw 27 july 2000 : 29 mar 2001
//winCE does not have StackWalk (or imagehlp.dll or dbghelp.dll) so we
//use the walkStack defined in process.C

//
// process::walkStack
//
// Walk the stack of thread 0 of our process.
// We assume that a trace is complete if we get to a main function
// (main, WinMain, or some variant of mainCRTStartup).
// If we can't get a complete trace, we return an empty vector.
//
// Note that this routine does *not* handle stack walks for multi-threaded
// programs - it only handles stack walks in thread zero.
// When we support multithreaded apps under Windows, this routine could
// take a thread parameter that specifies which thread to use.
//
// Also, note that we have *not* been able to find a mechanism that 
// can perform stack walks reliably for apps compiled with the FPO 
// (frame pointer omission) optimization.  Even if the FPO data is
// available, the Win32 StackWalk call either skips functions or bottoms
// out the stack too early.  Unfortunately, this means that we require
// that the app under study *not* use the FPO optimization.  (Specify /Oy- 
// for the VC++ compiler to turn this optimization off.)
//
vector<Address>
process::walkStack( bool noPause )
{
    bool needToCont = noPause ? false : (status() == running);
    vector<Address> pcs;

#ifndef BPATCH_LIBRARY
    startTimingStackwalk();
#endif

#ifdef DEBUG_STACKWALK
    cout << "\n<stack>" << endl;
#endif // DEBUG_STACKWALK

    // pause the application if necessary
    if( !noPause && !pause() )
    {
        // pause failed...give up
        cerr << "walkStack: pause failed" << endl;

#ifndef BPATCH_LIBRARY
	stopTimingStackwalk();
#endif
        return pcs;
    }

    // establish the current execution context
    CONTEXT cont;
	HANDLE hThread = (HANDLE)threads[0]->get_handle();

    cont.ContextFlags = CONTEXT_FULL;
    if (!GetThreadContext(hThread, &cont))
    {
        cerr << "walkStack: GetThreadContext failed:" << GetLastError() << endl;

#ifndef BPATCH_LIBRARY
	stopTimingStackwalk();
#endif
        return pcs;
    }

    STACKFRAME sf;
    ZeroMemory( &sf, sizeof(STACKFRAME) );
    sf.AddrPC.Offset = cont.Eip;
    sf.AddrPC.Mode = AddrModeFlat;
    sf.AddrFrame.Offset = cont.Ebp;
    sf.AddrFrame.Mode = AddrModeFlat;
    sf.AddrStack.Offset = cont.Esp;
    sf.AddrStack.Mode = AddrModeFlat;
    sf.AddrReturn.Mode = AddrModeFlat;

    // walk the stack, frame by frame
	// we use the Win32 StackWalk function to automatically
	// handle compiler optimizations, especially FPO optimizations
	bool done = false;
	bool reachedMain = false;
	while( !done )
	{
		STACKFRAME saved_sf = sf;
		BOOL walked;
		ADDRESS patchedAddrReturn;
		ADDRESS patchedAddrPC;
		instPoint* ip = NULL;
		function_base* fp = NULL;

		// set defaults for return address and PC
		patchedAddrReturn = saved_sf.AddrReturn;
		patchedAddrPC = saved_sf.AddrPC;

		// try to step through the stack using the current information
		walked = walkStackFrame( (HANDLE)getProcFileDescriptor(), hThread, &sf );

		if( !walked && (GetLastError() == ERROR_INVALID_ADDRESS) )
		{
			// try to patch the return address, in case it is outside
			// of the original text of the process.  It might be outside
			// the original text if the function has been relocated, or
			// if it represents a return from a call instruction that
			// has been relocated to a base tramp.
			//
			// we first try to patch the return address only, because
			// it is most likely that the return address only is out of
			// the original text space.  Once we process the first stack
			// frame, it appears that the StackWalk function transfers the
			// return address from the STACKFRAME struct to the PC, so 
			// once we've gotten a valid return address in a STACKFRAME,
			// the PC will be within the original text after the StackWalk
			sf = saved_sf;
            fp = findFunctionFromAddress( this, sf.AddrReturn.Offset );
			if( fp != NULL )
			{
				// because StackWalk seems to support it, we simply use 
				// the address of the function itself rather than
				// trying to do the much more difficult task of finding
				// the original address of the relocated instruction
				patchedAddrReturn.Offset = fp->addr();
				sf.AddrReturn = patchedAddrReturn;
			}

			// retry the stack step
			walked = walkStackFrame( (HANDLE)getProcFileDescriptor(), hThread, &sf );
		}

		if( !walked && (GetLastError() == ERROR_INVALID_ADDRESS) )
		{
			// patching the return address alone didn't work.
			//
			// try patching the return address and the PC
			sf = saved_sf;
            fp = findFunctionFromAddress( this, sf.AddrPC.Offset );
			if( fp != NULL )
			{
				// because StackWalk seems to support it, we simply use 
				// the address of the function itself rather than
				// trying to do the much more difficult task of finding
				// the original address of the relocated instruction
				patchedAddrPC.Offset = fp->addr();
				sf.AddrPC = patchedAddrPC;
			}

			// use the patched return address we calculated above
			sf.AddrReturn = patchedAddrReturn;

			// retry the stack step
			walked = walkStackFrame( (HANDLE)getProcFileDescriptor(), hThread, &sf );
		}

		if( !walked && (GetLastError() == ERROR_INVALID_ADDRESS) )
		{
			// patching both addresses didn't work
			//
			// try patching the PC only
			sf = saved_sf;
			sf.AddrPC = patchedAddrPC;

			// retry the stack step
			walked = walkStackFrame( (HANDLE)getProcFileDescriptor(), hThread, &sf );
		}


		// by now we've tried all of our tricks to handle the stack frame
		// if we haven't succeeded by now, we're not going to be able to
		// handle this frame
		if( walked )
		{
            Address pc = NULL;

			// save the PC for this stack frame
            // make sure we use the original address in case
            // it was outside the original text of the process
            if( saved_sf.AddrReturn.Offset == 0 )
            {
                // this was the first stack frame, so we had better
                // use the original PC
                pc = saved_sf.AddrPC.Offset;
            }
            else
            {
                // this was not the first stack frame
                // use the original return address
                pc = saved_sf.AddrReturn.Offset;
            }
            pcs.push_back(pc);

			// check whether we reached a known "main" function
			// we consider it a complete stack walk iff we see
			// a "main" function on the stack
			//
			// note: this criteria is not likely to be generally applicable -
			// it will fail for walking stacks in non-primary threads
			//
            fp = findFunctionFromAddress( this, pc );
			if( (fp != NULL) && 
				((fp == getMainFunction()) || 
                 (fp->prettyName() == "mainCRTStartup") ||
                 (fp->prettyName() == "wmainCRTStartup")) )
			{
				reachedMain = true;
			}

#ifdef DEBUG_STACKWALK
            cout << "0x" << setw(8) << setfill('0') << hex << pc << ": ";
            if( fp != NULL )
            {
                cout << fp->prettyName();
            }
            else
            {
                cout << "<unknown>";
            }
            if( sf.AddrPC.Offset != pc )
            {
                cout << " (originally 0x" << setw(8) << setfill('0') << hex << sf.AddrPC.Offset << ")";
            }
            cout << endl;
#endif // DEBUG_STACKWALK
		}
		else
		{
			// we tried everything we know to recover - we'll have to fail
			done = true;
		}
	}


    // handle incomplete stack walks
    // (defined as a stack on which the main function did not appear)
    if( !reachedMain )
    {
        // error - incomplete trace, return an empty vector
        pcs.resize( 0 );
    }
    else
    {
        pcs.push_back(NULL);
    }

    // resume the application if needed
    if( !noPause && needToCont )
    {
        if( !continueProc() )
        {
            cerr << "walkStack: continueProc failed" << endl;
        }
    }

#ifndef BPATCH_LIBRARY
        stopTimingStackwalk();
#endif
	return pcs;
}
#endif



/* 
   Loading libDyninstRT.dll

   We load libDyninstRT.dll dynamically, by inserting code into the
   application to call LoadLibraryA. We don't use the inferior RPC
   mechanism from class process because it already assumes that
   libdyninst is loaded (it uses the inferior heap).
   Instead, we use a simple inferior call mechanism defined below
   to insert the code to call LoadLibraryA("libdyninstRT.dll").
 */

Address loadDyninstDll(process *p, char Buffer[LOAD_DYNINST_BUF_SIZE]) {
    Address codeBase = p->getImage()->codeOffset();
	Address hackLoadLibAddr = 0x01f9ac30; //ccw 2 feb 2001 : HACK

    Address LoadLibBase;
    Address LoadLibAddr;
    Symbol sym;
 	//ccw 22 aug 2000 : this function hand codes x86 op codes!!!!
	//change these to mips instructions!!!

#ifdef mips_unknown_ce2_11 //ccw 14 aug 2000 : 29 mar 2001

	//ccw 2 feb 2001
	//this is a terrible hack.  i know that LoadLibrary is 0xac30 bytes from
	//the base address of coredll.dll from emprical evidence.  i have no
	//debug symbols for coredll.dll to prove this dynamically though....
	
	for(int i=0; i< (p->sharedObjects())->size();i++){
		
		if((*p->sharedObjects())[i]->getName() == "coredll.dll" ){
			hackLoadLibAddr = (*p->sharedObjects())[i]->getBaseAddress() + 0x0000bf1c;
			//0x0000ac30; //HACK 24 apr 2001
		}
	}

#if 0 
	//ccw 2 feb 2001 : this is taken out because the above
	//loop determines the correct address for LoadLibrary
	//this statement requires LoadLibrary to show up in
	//the map file of the mutatee, which only happens if
	//the mutatee explicitly calls LoadLibrary
	if (!p->getSymbolInfo("LoadLibraryW", sym, LoadLibBase)) {

	 if( !p->getSymbolInfo( "__imp_LoadLibraryW", sym, LoadLibBase )) //ccw 1 oct 2000, not a good idea
         {
	        printf("unable to find function LoadLibrary\n");
	        assert(0);
        }
    }
#endif 

#else
	if (!p->getSymbolInfo("_LoadLibraryA@4", sym, LoadLibBase)) {
        if( !p->getSymbolInfo( "_LoadLibraryA", sym, LoadLibBase ))
        {
	        printf("unable to find function LoadLibrary\n");
	        assert(0);
        }
    }
#endif

#ifndef mips_unknown_ce2_11 //ccw 2 feb 2001
    LoadLibAddr = sym.addr() + LoadLibBase;
    assert(LoadLibAddr);
#endif

    char ibuf[LOAD_DYNINST_BUF_SIZE];
    char *iptr = ibuf;
	memset(ibuf, '\0', LOAD_DYNINST_BUF_SIZE);//ccw 25 aug 2000

#ifdef mips_unknown_ce2_11 //ccw 22 aug 2000
	//the following are the instructions to load libdyninstRT_API.dll
	
	//ccw 2 oct 2000
	//the instructions we need to generate are:
	// lui $a0,hi16(codeBase)  # codeBase contains libBuf
	// ori $a0,lo16(codeBase)
	// lui $t0, hi16(hackLoadLibAddr)
	// ori $t0, lo16(hackLoadLibAddr)
	// jalr $t0
	// nop
	// nop
	// break
	// nop
	// nop
	// nop
	// where libBuf is the location of the string containing the name of the dll
	// in the remote process's memory

	WCHAR libBuf[] = L"\\libdyninstAPI_RT.dll";
	BYTE instBuff[4]; //buffer for the instruction to write to memory;

	//lay the name of the DLL to load into memory, UNICODE!
	assert(memcpy(iptr, (char*)libBuf, 42));
	iptr +=42;
	memset(iptr, '\0', 2);
	iptr +=2;

	//LUI //ccw 2 oct 2000
	instBuff[0]=((char*) &codeBase)[2];
	instBuff[1]=((char*) &codeBase)[3];
	instBuff[2]=0x04;
	instBuff[3]=0x3C;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//ORI //ccw 2 oct 2000
	instBuff[0]=((char*) &codeBase)[0];
	instBuff[1]=((char*) &codeBase)[1];
	instBuff[2]=0x84;
	instBuff[3]=0x34;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//LUI //ccw 2 feb 2001
	instBuff[0]=((char*) &hackLoadLibAddr)[2];
	instBuff[1]=((char*) &hackLoadLibAddr)[3];
	instBuff[2]=0x08;
	instBuff[3]=0x3C;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//ORI //ccw 2 feb 2001
	instBuff[0]=((char*) &hackLoadLibAddr)[0];
	instBuff[1]=((char*) &hackLoadLibAddr)[1];
	instBuff[2]=0x08;
	instBuff[3]=0x35;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//JALR
	instBuff[0]=0x09;
	instBuff[1]=0xf8;
	instBuff[2]=0x00;
	instBuff[3]=0x01;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;


	//NOP
	instBuff[0]=0x00;
	instBuff[1]=0x00;
	instBuff[2]=0x00;
	instBuff[3]=0x00;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	//NOP
	instBuff[0]=0x00;
	instBuff[1]=0x00;
	instBuff[2]=0x00;
	instBuff[3]=0x00;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;


	//DebugBreak();
	instBuff[0]=0x0D;
	instBuff[1]=0x00;
	instBuff[2]=0x00;
	instBuff[3]=0x00;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;


	//ccw 25 aug 2000 : throw some NOPs after the debugBreak just in case!
	//NOP
	instBuff[0]=0x00;
	instBuff[1]=0x00;
	instBuff[2]=0x00;
	instBuff[3]=0x00;
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;
	
	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

	memcpy(iptr, (char*) &instBuff, 4);
	iptr +=4;

#else

    // push nameAddr ; 5 bytes
    *iptr++ = (char)0x68; 
    *(int *)iptr = codeBase + 14; 
    iptr += sizeof(int);

    // call LoadLibrary ; 5 bytes
    *iptr++ = (char)0xe8;
    // offset relative to next instruction (address codeBase+10)
    *(int *)iptr = LoadLibAddr - (codeBase + 10);
    iptr += sizeof(int);

    // add sp, 4
    *iptr++ = (char)0x83; *iptr++ = (char)0xc4; *iptr++ = (char)0x04;

    // int3
    *iptr++ = (char)0xcc;

    if (!process::dyninstName.length())
        // check for an environment variable
#ifdef BPATCH_LIBRARY  // dyninstAPI loads a different run-time library
        process::dyninstName = getenv("DYNINSTAPI_RT_LIB");
#else
        process::dyninstName = getenv("PARADYN_LIB");
#endif

    if (!process::dyninstName.length())
        // if environment variable unset, use the default name/strategy
#ifdef BPATCH_LIBRARY
        process::dyninstName = "libdyninstAPI_RT.dll";
#else
        process::dyninstName = "libdyninstRT.dll";
#endif
	
    // make sure that directory separators are what LoadLibrary expects
    strcpy(iptr, process::dyninstName.string_of());
    for (unsigned int i=0; i<strlen(iptr); i++)
        if (iptr[i]=='/') iptr[i]='\\';
#endif

    p->readDataSpace((void *)codeBase, LOAD_DYNINST_BUF_SIZE, Buffer, false);
    p->writeDataSpace((void *)codeBase, LOAD_DYNINST_BUF_SIZE, ibuf);

#ifdef mips_unknown_ce2_11 //ccw 22 aug 2000
	//add to process, then to remoteDevice
    return codeBase+44;//this is for the offset of the string!
#else
	return codeBase;
#endif
}



// osTraceMe is not needed in Windows NT
void OS::osTraceMe(void) {}


#ifndef BPATCH_LIBRARY
void checkProcStatus() {
    int wait_status;
    extern void doDeferredRPCs();

    process::waitProcs(&wait_status);
    doDeferredRPCs();
}
#endif


// ccw 2 may 2001 win2k fixes
// when you launch a process to be debugged with win2k (as in createProcess)
// the system sends you back at least two debug events before starting the
// process.  a debug event is also sent back for every dll that is loaded
// prior to starting main(), these include ntdll.dll and kernel32.dll and any
// other dlls the process needs that are not loaded with an explicit call
// to LoadLibrary().
//
// dyninst catches the first debug event (CREATE_PROCESS) and initializes
// various process specific data structures.
// dyninst catches the second debug event (an EXCEPTION_DEBUG_EVENT) and used this
// event as a trigger to put in the bit of code that forced the mutatee to
// load libdyninstAPI_RT.dll.  In win2k, this does not work.  The bit of code
// is run and the trailing DebugBreak is caught and handled but the Dll will not
// be loaded.  The EXCEPTION_DEBUG_EVENT must be handled and continued from before
// LoadLibrary will perform correctly.
//
// the fix for this is to handle this EXCEPTION_DEBUG_EVENT, and put a DebugBreak (0xcc) 
// at the beginning of main() in the mutatee.  catching that DebugBreak allows dyninst
// to write in the bit of code used to load the libdyninstAPI_RT.dll.
//
// after this, dyninst previously instrumented the mutatee to force the execution of
// DYNINSTinit() in the dll.  in order to take out this bit of complexity, the DllMain()
// function in the dll, which is run upon loading the dll, is used to automatically call
// DYNINSTinit().
//
// DYNINSTinit() takes two parameters, a flag denoting how dyninst attached to this process
// and the pid of the mutator.  These are passed from the mutator to the mutatee by
// finding a variable in the dll and writing the correct values into the mutatee's
// address space.  When a Dll is loaded, a LOAD_DLL debug event is thrown before the
// execution of DllMain(), so dyninst catches this event, writes the necessary values
// into the mutatee memory, then lets DllMain() call DYNINSTinit().
// the DebugBreak() at the end of DYNINSTinit() is now removed for NT/win2K
//
// the bit of code inserted to load the dll fires a DebugBreak() to signal that it is
// done. dyninst catches this, patches up the code that was used to load the dll, 
// replaces what was overwritten in main() and resets the instruction pointer (EIP)
// to the beginning of main().

int secondBkpt = 0; //ccw 30 apr 2001 : used to signal that we have seen the first EXCEPTION_DEBUG_EVENT
int mungeAddr = 0; //ccw 2 may 2001 : used to track where we wrote the LoadLibrary code
byte savedOpCode[256]; //ccw 2 may 2001 : the op code we overwrite in main()
byte newOpCode[256];   // ccw 27 june 2001 : the op code we add in main()
//ccw 6 july 2001 the two above arrays are 256 bytes because writing one byte does not always cause the
//instruction cache to be reloaded. 

Address mainAddr=0 ; //ccw 2 may 2001 : the address of main()

//ccw 2 may 2001 : the following function sets the varible name to the value given.
// returns true if it is successful.
bool setVariable(process* p, string name, int value){
    	string full_name = string("_") + string(name);
	Symbol syminfo;
	if (!p->getSymbolInfo(full_name, syminfo)) {
		string short_name(name);
		if (!p->getSymbolInfo(short_name, syminfo)) {
		    assert(0);
		}
	}

	if( syminfo.type() == Symbol::PDST_FUNCTION)
   		assert(0); 
	Address tmpAddr = syminfo.addr();
	return p->writeDataSpace((void*)tmpAddr, sizeof(int), (void*) &value);
}

#ifdef mips_unknown_ce2_11
int secondDLL = 0; //ccw 24 oct 2000 : 28 mar 2001
#endif
DWORD continueType = DBG_CONTINUE; //ccw 25 oct 2000 : 28 mar 2001

#ifdef i386_unknown_nt4_0 //ccw 21 july 2001

/*
   wait for inferior processes to terminate or stop.
*/
#ifdef BPATCH_LIBRARY
int process::waitProcs(int *status, bool block) {
#else
int process::waitProcs(int *status) {
#endif
    DEBUG_EVENT debugEv;
    process *p;
#ifdef BPATCH_LIBRARY
    *status = 0;
#endif

    // We wait for 1 millisecond here. On the Unix platforms, the wait
    // happens on the select in controllerMainLoop. But on NT, because
    // we have to handle traps, we set the timeout on the select as 0,
    // so that we can check for traps quickly
#ifdef BPATCH_LIBRARY
    DWORD milliseconds;
    if (block) milliseconds = INFINITE;
    else milliseconds = 1;
    if (!WaitForDebugEvent(&debugEv, milliseconds))
#else

    if (!WaitForDebugEvent(&debugEv, 1))

#endif
	return 0;

    //printf("Debug event from process %d, tid %d\n", debugEv.dwProcessId,
    //	 debugEv.dwThreadId);
    p = findProcess(debugEv.dwProcessId);
    if (p == NULL) {
	/* 
	   this case can happen when we create a process, but then are
	   unable to parse the symbol table, and so don't complete the
	   creation of the process. We just ignore the event here.  
	*/
		ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			   DBG_CONTINUE);
	return 0; 
    }

    switch (debugEv.dwDebugEventCode) {
    case EXCEPTION_DEBUG_EVENT: {
	DWORD exnCode = debugEv.u.Exception.ExceptionRecord.ExceptionCode;
	switch(exnCode) {
	case EXCEPTION_BREAKPOINT: {
	    //printf("Debug breakpoint exception, %d, addr = %x\n", 
	    //       debugEv.u.Exception.ExceptionRecord.ExceptionFlags,
	    //       debugEv.u.Exception.ExceptionRecord.ExceptionAddress);

	    	if(!secondBkpt && !p->wasCreatedViaAttach()){
			// if this was created via attach then
			// we are not at main() so no reason to
			// put a breakpoint there, we are already
			// at a state that will allow loadLibrary() to
			// succeed. 
			//ccw this is the breakpoint BEFORE main();
			//i need to insert the 0xcc into main() now.
			//byte debugOpCode = 0xcc;
	 	    	p->savedRegs = p->getRegisters();
			function_base *mainFunc;
			//DebugBreak();//ccw 14 may 2001  
			if (!((mainFunc = p->findOneFunction("main")))){
				if(!(mainFunc = p->findOneFunction("_main"))){
					if(!(mainFunc = p->findOneFunction("WinMain"))){
						if(!(mainFunc = p->findOneFunction("_WinMain"))){
							assert(0);
						}
					}
				}

			}
			mainAddr = mainFunc->addr();
			//ccw save what we overwrite for later.
			p->readDataSpace_((void*) (mainAddr), 256, (void*)savedOpCode);

			// insert the breakpoint into the code we just read
			CopyMemory( newOpCode, savedOpCode, 256 );
			newOpCode[0] = 0xcc;

			// write the modified code sequence back
			p->writeDataSpace((void*) (mainAddr), 256, (void*)newOpCode);
			p->flushInstructionCache_( (void*)mainAddr, 256 );

			secondBkpt=1;
			break;
		}



	    if (!p->reachedFirstBreak) {
		int inst_offset =0xd; 
		if (!p->hasLoadedDyninstLib && !p->isLoadingDyninstLib) {
			//ccw 11 july 2001: IS THIS TRUE?
			//ccw 28 feb 2001
			//for MIPS/WINCE this never happens here, it is done in the
			//LOAD_DLL exception handler.  this is because of the
			//bug in which WINCE only passes one debug break when
			//it starts a process (unlike NT which gives two).
			//more explaination below in LOAD_DLL
			//DebugBreak();//ccw 30 jan 2001
			continueType = DBG_CONTINUE;//ccw 25 oct 2000 : 29 mar 2001



		    Address addr = loadDyninstDll(p, p->savedData);
	   	    mungeAddr = addr; //ccw 15 june 2001
		    p->savedRegs = p->getRegisters();
		    p->changePC(addr);
		    //p->LoadDyninstTrapAddr = addr + 0xd;
		    p->isLoadingDyninstLib = true;
		    break;
		} else if (p->isLoadingDyninstLib
			&& ((int) mungeAddr + inst_offset) ==
			(int) debugEv.u.Exception.ExceptionRecord.ExceptionAddress) {

			//ccw 2 may 2001 the above test is changed so that we only
			// respond to events after
			// the load library code has completed.  (we know that the 
			// DebugBreak 0xd bytes from 
			// the beginning of where we inserted code signals that we
			// have finished load library) 

		    p->isLoadingDyninstLib = false;
		    p->hasLoadedDyninstLib = true;
	 		//ccw 25 june 2001 NEED TO FLUSH ICACHE here
			p->flushInstructionCache_((void*) ((w32CONTEXT*) p->savedRegs)->Eip,0x100);


			if(mainAddr){
			    // patch main() back to its original form
			    p->writeDataSpace((void*) (mainAddr),16,(void*) savedOpCode); //ccw 2 may 2001
			    // reset the IP to run what we just inserted.
			    ((w32CONTEXT*) p->savedRegs)->Eip-=1; //ccw 2 may 2001
			}

		    p->restoreRegisters(p->savedRegs);
		    delete p->savedRegs;
		    p->writeDataSpace((void *)p->getImage()->codeOffset(),
				      LOAD_DYNINST_BUF_SIZE,
				      (void *)p->savedData);
			if(mainAddr){
				//ccw 25 june 2001 NEED TO FLUSH ICACHE here
				//p->flushInstructionCache_((void*) ((CONTEXT*) p->savedRegs)->Eip,0x100);
				p->readDataSpace_((void*)(mainAddr), 256, (void*) newOpCode);
				p->writeDataSpace((void*) (mainAddr),256,(void*) savedOpCode); //ccw 2 may 2001
				//technically, we only need to writeDataSpace once, up at the previous
				// if statement. BUT writing once does not seem to always flush the 
				// icache, and neither does FlushInsturctionCache()
				// the readDataSpace_ is completely spurrious but is here for the same
				// reason, to get the icache updated correctly with the old op code before
				// we continue. 
			}
	
	
		}else{
			break;  //ccw 20 june 2001 this is the DYNINSTinit breakpoint
		}
		p->reachedFirstBreak = true;
      
		if (!p->createdViaAttach) {
		    int pid = p->getPid();
		    p->status_ = stopped;

		    if (!p->initDyninstLib()) {
			OS::osKill(pid);
			handleProcessExit(p, -1);
#ifdef BPATCH_LIBRARY
			*status = SIGEM_SIGNALED | SIGTERM;
			return p->getPid();
#else
			return 0;
#endif
		    }

		    string buffer = string("PID=") + string(pid);
		    buffer += string(", passed trap at start of program");
		    statusLine(buffer.string_of());
      
		    buffer=string("PID=") + string(pid) 
			 + ", installing call to DYNINSTinit()";
		    statusLine(buffer.string_of());
#if !defined(BPATCH_LIBRARY)
		    p->installBootstrapInst();//ccw 20 june 2001 
			// now we dont need to do this,  DLLMain() does it.
     			//this still needs to be done for paradyn
 
		    // now, let main() and then DYNINSTinit() get invoked.  As it
		    // completes, DYNINSTinit() does a DYNINSTbreakPoint, at which time
		    // we read bootstrap information "sent back" (in a global vrble),
		    // propagate metric instances, do tp->newProgramCallbackFunc(), etc.
		     break; //ccw 20 june 2001, commented out  with p->installBootstrapInst()
#endif
		} else {
		    //printf("First breakpoint after attach\n");
		    p->pause_();
		    break;
		}
	    }

	    // the process has already been initialized
	    // First lookup in the tramp table, to see if the breakpoint
	    // is from an instrumentation point.

	    // lookup in tramp table


	    Address trampAddr = 0;
	    unsigned u; unsigned k;
	    unsigned key = 
		(unsigned)debugEv.u.Exception.ExceptionRecord.ExceptionAddress;
	    for (u = HASH1(key); 1; u = (u + HASH2(key)) % TRAMPTABLESZ) {
		k = p->trampTable[u].key;
		if (k == 0)
		    break;
		else if (k == key) {
		    trampAddr = p->trampTable[u].val;
		    break;
		}
	    }
	    if (trampAddr) {
		    // this is a trap from an instrumentation point

            // find the current thread
            pdThread* currThread = NULL;
		    for(unsigned int i = 0; i < p->threads.size(); i++)
            {
		        if ((unsigned)p->threads[i]->get_tid() == debugEv.dwThreadId)
                {
                    currThread = p->threads[i];
                    break;
                }
            }
            assert( currThread != NULL );

            // Due to a race between the processing of trap debug events
            // and the desire to run an inferior RPC, it is possible that
            // we hit the trap, set ourselves up to run an inferior RPC,
            // and then processed the trap notification.  If this happens,
            // we don't end up running *any* of the inferior RPC code.
            //
            // We can tell that this is what's happened based on the
            // thread's Eip - if it doesn't match the ExceptionAddress, we
            // know that we had tried to reset the Eip to execute an inferior
            // RPC.  In that case, we leave the Eip alone here, which means
            // the inferior RPC code will execute when we continue the thread.
            // We have to remember that we need to execute this instrumentatin
            // once the inferior RPC is done, however.
            // 
			CONTEXT cont;
			cont.ContextFlags = CONTEXT_FULL;
			if( !GetThreadContext( (HANDLE)currThread->get_handle(), &cont ) )
            {
			    assert(false);
            }
            if( debugEv.u.Exception.ExceptionRecord.ExceptionAddress == (PVOID)(cont.Eip - 1) )
            {
                // The Eip indicates we've just executed a trap instruction
                // reset the Eip to the address of the base tramp
                cont.Eip = trampAddr;
                if( !SetThreadContext( (HANDLE)currThread->get_handle(), &cont ) )
                {
                    assert( false );
                }
            }
            else
            {
                // We leave the Eip alone, since we've set it to execute 
                // at inferior RPC code.  However, we need to remember that
                // we should execute the base tramp once the inferiorRPC is
                // completed.
                currThread->set_pending_tramp_addr( trampAddr );
            }

		    break;
	    }

	    // If it is not from an instrumentation point,
	    // it could be from a call to DYNINSTbreakPoint
	    // or an inferior procedure call

	    p->status_ = stopped;
	    p->pause_();
	    int result = p->procStopFromDYNINSTinit();
	    assert(result >= 0 && result <= 2);
	    if (result != 0) {
		if (result == 1) {
	    		//p->pause_();
		} else {
		    // procStopFromDYNINSTinit called continueProc()
		    // What should we do here?
		    break;
		}
	    } else if (p->handleTrapIfDueToRPC()) {
		//p->pause_();
		break;
	    } else {
#ifdef BPATCH_LIBRARY /* Don't ignore unknown bkpt in library; leave stopped. */
		*status = SIGEM_STOPPED | SIGTRAP;

#else
		// Unknown breakpoint: we just ignore it
		printf("Debug breakpoint exception, %d, addr = %x\n", 
		       debugEv.u.Exception.ExceptionRecord.ExceptionFlags,
		       debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
		p->status_ = running;
		p->continueProc_();
#endif
	    }
	    break;
	}
	case EXCEPTION_ILLEGAL_INSTRUCTION:
	    //printf("Illegal instruction\n");
	    p->pause_();
	    p->status_ = stopped;
	    if( p->handleTrapIfDueToRPC() )
        {
		    // handleTrapIfDueToRPC calls continueProc()
            // however, under Windows NT, it doesn't actually 
            // continue the thread until the ContinueDbgEvent call is made

            // We take advantage of this fact to ensure that any
            // pending instrumentation is executed.  (I.e., instrumentation
            // put off so we could be sure to execute inferior RPC code is
            // now executed.)
            if( p->currRunningRPCs.size() == 0 )
            {
                // we finished the inferior RPC, so we can now execute
                // any pending instrumentation

                // find the current thread
                pdThread* currThread = NULL;
                for( unsigned int i = 0; i < p->threads.size(); i++ )
                {
		            if ((unsigned)p->threads[i]->get_tid() == debugEv.dwThreadId)
                    {
                        currThread = p->threads[i];
                        break;
                    }
                }
                assert( currThread != NULL );

                Address pendingTrampAddr = currThread->get_pending_tramp_addr();
                if( pendingTrampAddr != NULL )
                {
                    // reset the Eip to the pending instrumentation
                    CONTEXT ctxt;
                    ctxt.ContextFlags = CONTEXT_FULL;
                    if( !GetThreadContext( (HANDLE)currThread->get_handle(), &ctxt ) )
                    {
                        assert( false );
                    }
                    ctxt.Eip = pendingTrampAddr;
                    if( !SetThreadContext( (HANDLE)currThread->get_handle(), &ctxt ) )
                    {
                        assert( false );
                    }
                    currThread->set_pending_tramp_addr( NULL );
                }
            }
		    break;
	    }
	    p->status_ = running;
	    p->continueProc_();
	   ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
					   DBG_EXCEPTION_NOT_HANDLED);
	    return 0;

	case EXCEPTION_ACCESS_VIOLATION:
	    printf("Access violation exception, %d, addr = %08x\n", 
		   debugEv.u.Exception.ExceptionRecord.ExceptionFlags,
		   debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
	    dumpMem(p, debugEv.u.Exception.ExceptionRecord.ExceptionAddress, 32);

        {
           vector<Address> pcs = p->walkStack( false );
            for( unsigned i = 0; i < pcs.size(); i++ )
            {
                function_base* f = p->findFuncByAddr( pcs[i] );
                const char* szFuncName = (f != NULL) ? f->prettyName().string_of() : "<unknown>";
                fprintf( stderr, "%08x: %s\n", pcs[i], szFuncName );
            }
        }
	    //	ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
	    //			   DBG_EXCEPTION_NOT_HANDLED);
	    //	break;
	default:
	    //printf("exeption %x, addr %x\n", exnCode, 
	    //   debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
		ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
				   DBG_EXCEPTION_NOT_HANDLED);

	    return 0;
	}
    } break;

    case CREATE_THREAD_DEBUG_EVENT: {
	//printf("create thread, tid = %d\n", debugEv.dwThreadId);
	assert(p->threads.size() > 0); // main thread should be already defined
	pdThread *t = new pdThread(p, debugEv.dwThreadId,
				   debugEv.u.CreateThread.hThread);
	p->threads.push_back(t);
    } break;

    case CREATE_PROCESS_DEBUG_EVENT: {
	//CREATE_PROCESS_DEBUG_INFO info = debugEv.u.CreateProcessInfo;
	//printf("CREATE_PROCESS event: %d\n", debugEv.dwProcessId);
	p = findProcess(debugEv.dwProcessId);
	if (p) {
	    secondBkpt = 0; //ccw 27 june 2001 : allows for more than one mutatee to be started
	    //fprintf(stderr,"create process: base = %x\n", info.lpBaseOfImage);
	    if (p->threads.size() == 0) {
		// define the main thread
		p->threads.push_back(new pdThread(p));
	    }
	    p->threads[0]->update_handle(debugEv.dwThreadId, 
					 debugEv.u.CreateProcessInfo.hThread);
	    
	}
    } break;

    case EXIT_THREAD_DEBUG_EVENT: {
	//printf("exit thread, tid = %d\n", debugEv.dwThreadId);
	unsigned nThreads = p->threads.size();
	// start from one to skip main thread
	for (unsigned u = 1; u < nThreads; u++) {
	    if ((unsigned)p->threads[u]->get_tid() == debugEv.dwThreadId) {
		delete p->threads[u];
		p->threads[u] = p->threads[nThreads-1];
		p->threads.resize(nThreads-1);
		break;
	    }
	}
    } break;

    case EXIT_PROCESS_DEBUG_EVENT:
	p = findProcess(debugEv.dwProcessId);
	if (p) {
            char errorLine[1024];
            sprintf(errorLine, "Process %d has terminated with code 0x%x\n", 
                p->getPid(), debugEv.u.ExitProcess.dwExitCode);
            statusLine(errorLine);
            logLine(errorLine);
	    handleProcessExit(p, debugEv.u.ExitProcess.dwExitCode);
	}
#ifdef BPATCH_LIBRARY
	*status = SIGEM_EXITED;
#else
	break;
#endif
	break; //ccw 23 july 2001
    case LOAD_DLL_DEBUG_EVENT: {
	//printf("load dll: hFile=%x, base=%x, debugOff=%x, debugSz=%d lpname=%x, %d\n",
	//       debugEv.u.LoadDll.hFile, debugEv.u.LoadDll.lpBaseOfDll,
	//       debugEv.u.LoadDll.dwDebugInfoFileOffset,
	//       debugEv.u.LoadDll.nDebugInfoSize,
	//       debugEv.u.LoadDll.lpImageName,
	//       /*debugEv.u.LoadDll.fUnicode*/
	//       GetFileSize(debugEv.u.LoadDll.hFile,NULL));

	// We are currently not handling dynamic libraries loaded after the
	// application starts.
	if (p->isBootstrappedYet())
	  break;

	// set the proc handle for the process that is loading the library
	// This is need when we use the imagehelp library to get the
	// symbols for the dll.
	// kludgeProcHandle = (HANDLE)p->getProcFileDescriptor();

        void *addr = debugEv.u.LoadDll.lpImageName;
        int ptr = 0;
	char nameBuffer[MAX_PATH];
	if (addr) {
	    /***
	      addr is a pointer to the image name (in the application address space)
	      There is a problem here, because this pointer can be null and it is
	      null in some cases (in particular, if we attach to a running process
	      it is always null). This means that it can be difficult to find
	      the name of an image. 

	      Currently, if we can't find the name, we just ignore the
	      image.  However, we need to at least be able to parse
	      kernel32.dll, because this library contains some very
	      important functions, including LoadLibrary, which we
	      need to call to load libdyninstRT.dll. We use the following
	      kludge to find if a library is kernel32.dll: the function
	      klugde_isKernel32Dll open a handle to kernel32.dll. We then
	      find if the handle we get from the LOAD_DLL_DEBUG_EVENT
	      is for the same file.
	    ***/

	    p->readDataSpace_(addr, 4, &ptr);
	    //printf("ptr = %x\n", ptr);
	    string kernel32Path;
		// ****************
		//ccw 10 aug 2000 : what is going on here???? ***************
		//at this point ptr contains the first 4 bytes of the name of the
		//DLL!?!?!?! IS DEBUGEVENT LOADDLL diff on nt/ce? MAYBE
		// ****************

	    if (ptr) {

		if (debugEv.u.LoadDll.fUnicode) {
		    unsigned short wbuffer[128];
		    p->readDataSpace_((void *)ptr, 128*sizeof(short), wbuffer);
		    WideCharToMultiByte(CP_ACP, 0, wbuffer, 128,
					nameBuffer, 128, NULL, NULL);
		} else {
		    p->readDataSpace_((void *)ptr, 128, nameBuffer);
		}
		//printf("Dll name: %s, base = %x\n", nameBuffer, debugEv.u.LoadDll.lpBaseOfDll);
	    }
	    else if (kludge_isKernel32Dll(debugEv.u.LoadDll.hFile, kernel32Path)) {
	        assert(kernel32Path.length() > 0);
	        assert(kernel32Path.length() < sizeof(nameBuffer));
		strcpy(nameBuffer, kernel32Path.string_of());
	    } else
	        break;
	}

	// try to load symbols for the DLL
        if (!SymLoadModule((HANDLE)p->getProcFileDescriptor(),
			   debugEv.u.LoadDll.hFile,
			   NULL, NULL, 0, 0)) {

		char msgText[1024];

		sprintf( msgText, "SymLoadModule failed for %s: 0x%x\n",
			nameBuffer, GetLastError() );

		logLine(msgText);
	}

	// discover structure of new DLL, and incorporate into our
	// list of known DLLs
	if (addr) {

	    shared_object *so =
		new shared_object(string(nameBuffer), 0,false, true, true, 0);

	    assert(p->dyn);
	    p->dyn->sharedObjects.push_back(so);
	    if (!p->shared_objects) {
	      p->shared_objects = new vector<shared_object *>;
	    }
	    (*(p->shared_objects)).push_back(so);
#ifndef BPATCH_LIBRARY
	    tp->resourceBatchMode(true);
#endif 
	    p->addASharedObject(*so);
#ifndef BPATCH_LIBRARY
	    tp->resourceBatchMode(false);
#endif 
	    p->setDynamicLinking();
	  }
#ifdef BPATCH_LIBRARY
	for(int kk = 0;kk< MAX_PATH && nameBuffer[kk] != '\0';kk++){ //ccw 2 may 2001
		nameBuffer[kk] = toupper(nameBuffer[kk]);
	}
	if(strstr(nameBuffer, "LIBDYNINSTAPI_RT.DLL")){
	// ccw 2 may 2001
	//when we load libdyninstAPI_RT.dll we need to find the local
	//variables in the dll that will hold the cause and pid so when
	//DLLMain is called they will be correctly passed to DYNINSTinit
		if(!setVariable(p, "libdyninstAPI_RT_DLL_localCause", 1)){
			assert(0);
		}
		if(!setVariable(p, "libdyninstAPI_RT_DLL_localPid", getpid())){
			assert(0);
		}
	}
#endif	
	}
    break;
    case UNLOAD_DLL_DEBUG_EVENT:
	// TODO
	//printf("unload dll\n");
	break;
    case OUTPUT_DEBUG_STRING_EVENT:

	//printf("output debug string\n");
	break;
    case RIP_EVENT:
	//printf("rip event\n");
	break;
    default:
	;
	//printf("unknown debug event\n");
	
    }
    if (!ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			    DBG_CONTINUE)) {
	DebugBreak();
	printf("ContinueDebugEvent failed\n");
	printSysError(GetLastError());
    }

#ifdef BPATCH_LIBRARY
    if (*status)
	return debugEv.dwProcessId;
    else
	return 0;
#else
    return 0;
#endif
}

#elif mips_unknown_ce2_11 //ccw 21 july 2001


#ifdef BPATCH_LIBRARY
int process::waitProcs(int *status, bool block) {
#else
int process::waitProcs(int *status) {
#endif
    DEBUG_EVENT debugEv;
    process *p;
#ifdef BPATCH_LIBRARY
    *status = 0;
#endif

    // We wait for 1 millisecond here. On the Unix platforms, the wait
    // happens on the select in controllerMainLoop. But on NT, because
    // we have to handle traps, we set the timeout on the select as 0,
    // so that we can check for traps quickly
#ifdef BPATCH_LIBRARY
    DWORD milliseconds;
    if (block) milliseconds = INFINITE;
    else milliseconds = 1;
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteWaitForDebugEvent(&debugEv, milliseconds))
#else
    if (!WaitForDebugEvent(&debugEv, milliseconds))
#endif
	return 0;
#else

#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteWaitForDebugEvent(&debugEv, 1))
#else
    if (!WaitForDebugEvent(&debugEv, 1))
#endif
	return 0;
#endif

    //printf("Debug event from process %d, tid %d\n", debugEv.dwProcessId,
    //	 debugEv.dwThreadId);

    p = findProcess(debugEv.dwProcessId);
    if (p == NULL) {
	/* 
	   this case can happen when we create a process, but then are
	   unable to parse the symbol table, and so don't complete the
	   creation of the process. We just ignore the event here.  
	*/
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
		BPatch::bpatch->rDevice->RemoteContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			   DBG_CONTINUE);
#else
		ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			   DBG_CONTINUE);

#endif
	return 0; 
    }

	continueType = DBG_CONTINUE;//ccw 25 oct 2000 : 29 mar 2001
    switch (debugEv.dwDebugEventCode) {
    case EXCEPTION_DEBUG_EVENT: {
	DWORD exnCode = debugEv.u.Exception.ExceptionRecord.ExceptionCode;
	switch(exnCode) {
	case EXCEPTION_BREAKPOINT: {
	    //printf("Debug breakpoint exception, %d, addr = %x\n", 
	    //       debugEv.u.Exception.ExceptionRecord.ExceptionFlags,
	    //       debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
#ifdef mips_unknown_ce2_11 //ccw 20 mar 2001 : 29 mar 2001
		continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 25 oct 2000
#endif

	    if (!p->reachedFirstBreak) {

		if (!p->hasLoadedDyninstLib && !p->isLoadingDyninstLib) {
			//ccw 28 feb 2001
			//for MIPS/WINCE this never happens here, it is done in the
			//LOAD_DLL exception handler.  this is because of the
			//bug in which WINCE only passes one debug break when
			//it starts a process (unlike NT which gives two).
			//more explaination below in LOAD_DLL
			//DebugBreak();//ccw 30 jan 2001
			continueType = DBG_CONTINUE;//ccw 25 oct 2000 : 29 mar 2001

		    Address addr = loadDyninstDll(p, p->savedData);
		    p->savedRegs = p->getRegisters();
		    p->changePC(addr);
		    //p->LoadDyninstTrapAddr = addr + 0xd;
		    p->isLoadingDyninstLib = true;
		    break;
		} else if (p->isLoadingDyninstLib) {
#ifdef mips_unknown_ce2_11 //ccw 20 mar 2001 : 29 mar 2001
			continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 25 oct 2000
#endif
		    p->isLoadingDyninstLib = false;
		    p->hasLoadedDyninstLib = true;
		    p->restoreRegisters(p->savedRegs);
		    delete p->savedRegs;
		    p->writeDataSpace((void *)p->getImage()->codeOffset(),
				      LOAD_DYNINST_BUF_SIZE,
				      (void *)p->savedData);
		}

		p->reachedFirstBreak = true;
      
		if (!p->createdViaAttach) {
		    int pid = p->getPid();
		    p->status_ = stopped;

		    if (!p->initDyninstLib()) {
			OS::osKill(pid);
			handleProcessExit(p, -1);
#ifdef BPATCH_LIBRARY
			*status = SIGEM_SIGNALED | SIGTERM;
			return p->getPid();
#else
			return 0;
#endif
		    }

		    string buffer = string("PID=") + string(pid);
		    buffer += string(", passed trap at start of program");
		    statusLine(buffer.string_of());
      
		    buffer=string("PID=") + string(pid) 
			 + ", installing call to DYNINSTinit()";
		    statusLine(buffer.string_of());
		    p->installBootstrapInst();
      
		    // now, let main() and then DYNINSTinit() get invoked.  As it
		    // completes, DYNINSTinit() does a DYNINSTbreakPoint, at which time
		    // we read bootstrap information "sent back" (in a global vrble),
		    // propagate metric instances, do tp->newProgramCallbackFunc(), etc.
		    break;
		} else {
		    //printf("First breakpoint after attach\n");
		    p->pause_();
		    break;
		}
	    }

	    // the process has already been initialized
	    // First lookup in the tramp table, to see if the breakpoint
	    // is from an instrumentation point.

	    // lookup in tramp table

#ifndef mips_unknown_ce2_11 //ccw 26 july 2000 : 29 mar 2001
		//the following code checks to see if the breakpoint
		//was triggered by code put in by the mutator to force
		//the mutatee to stop so we can set the PC correctly
		//this is done when a point is not big enough to have
		//a jump put in it, hence the breakpoint bounces it to
		//here and then this code looks up where it should go
		//in the trampTable, a list of all addresses that have been 
		//instrumented and their (jump) targets.
		//MIPS does not have this problem as all the instructions
		//are the same size.

	    Address trampAddr = 0;
	    unsigned u; unsigned k;
	    unsigned key = 
		(unsigned)debugEv.u.Exception.ExceptionRecord.ExceptionAddress;
	    for (u = HASH1(key); 1; u = (u + HASH2(key)) % TRAMPTABLESZ) {
		k = p->trampTable[u].key;
		if (k == 0)
		    break;
		else if (k == key) {
		    trampAddr = p->trampTable[u].val;
		    break;
		}
	    }
	    if (trampAddr) {
		    // this is a trap from an instrumentation point

            // find the current thread
            pdThread* currThread = NULL;
		    for(unsigned int i = 0; i < p->threads.size(); i++)
            {
		        if ((unsigned)p->threads[i]->get_tid() == debugEv.dwThreadId)
                {
                    currThread = p->threads[i];
                    break;
                }
            }
            assert( currThread != NULL );

            // Due to a race between the processing of trap debug events
            // and the desire to run an inferior RPC, it is possible that
            // we hit the trap, set ourselves up to run an inferior RPC,
            // and then processed the trap notification.  If this happens,
            // we don't end up running *any* of the inferior RPC code.
            //
            // We can tell that this is what's happened based on the
            // thread's Eip - if it doesn't match the ExceptionAddress, we
            // know that we had tried to reset the Eip to execute an inferior
            // RPC.  In that case, we leave the Eip alone here, which means
            // the inferior RPC code will execute when we continue the thread.
            // We have to remember that we need to execute this instrumentation
            // once the inferior RPC is done, however.
            // 
			CONTEXT cont;
			cont.ContextFlags = CONTEXT_FULL;
			if( !GetThreadContext( (HANDLE)currThread->get_handle(), &cont ) )
            {
			    assert(false);
            }
            if( debugEv.u.Exception.ExceptionRecord.ExceptionAddress == (PVOID)(cont.Eip - 1) )
            {
                // The Eip indicates we've just executed a trap instruction
                // reset the Eip to the address of the base tramp
                cont.Eip = trampAddr;
                if( !SetThreadContext( (HANDLE)currThread->get_handle(), &cont ) )
                {
                    assert( false );
                }
            }
            else
            {
                // We leave the Eip alone, since we've set it to execute 
                // at inferior RPC code.  However, we need to remember that
                // we should execute the base tramp once the inferiorRPC is
                // completed.
                currThread->set_pending_tramp_addr( trampAddr );
            }

		    break;
	    }

#endif

	    // If it is not from an instrumentation point,
	    // it could be from a call to DYNINSTbreakPoint
	    // or an inferior procedure call

	    p->status_ = stopped;
	    p->pause_();
	    int result = p->procStopFromDYNINSTinit();
	    assert(result >= 0 && result <= 2);
	    if (result != 0) {
		if (result == 1) {
			// DBG_CONTINUE does not work!
#ifdef mips_unknown_ce2_11 //ccw 20 mar 2001 : 29 mar 2001
			continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 1 mar 2001
#endif
		    //p->pause_();
		} else {
		    // procStopFromDYNINSTinit called continueProc()
		    // What should we do here?
		    break;
		}
	    } else if (p->handleTrapIfDueToRPC()) {
		//p->pause_();
		break;
	    } else {
#ifdef BPATCH_LIBRARY /* Don't ignore unknown bkpt in library; leave stopped. */
		*status = SIGEM_STOPPED | SIGTRAP;
				//ccw 5 feb 2001 : below : 29 mar 2001
				//AARRRGGHHHH!!!
				// DBG_CONTINUE does not work!
#ifdef mips_unknown_ce2_11 //ccw 20 mar 2001 : 29 mar 2001
		continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 28 feb 2001
#endif

#else
		// Unknown breakpoint: we just ignore it
		printf("Debug breakpoint exception, %d, addr = %x\n", 
		       debugEv.u.Exception.ExceptionRecord.ExceptionFlags,
		       debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
		p->status_ = running;
		p->continueProc_();
#endif
	    }
	    break;
	}
	case EXCEPTION_ILLEGAL_INSTRUCTION:
	    //printf("Illegal instruction\n");
	    p->pause_();
	    p->status_ = stopped;
	    if( p->handleTrapIfDueToRPC() )
        {
		    // handleTrapIfDueToRPC calls continueProc()
            // however, under Windows NT, it doesn't actually 
            // continue the thread until the ContinueDbgEvent call is made

            // We take advantage of this fact to ensure that any
            // pending instrumentation is executed.  (I.e., instrumentation
            // put off so we could be sure to execute inferior RPC code is
            // now executed.)
            if( p->currRunningRPCs.size() == 0 )
            {
                // we finished the inferior RPC, so we can now execute
                // any pending instrumentation

                // find the current thread
                pdThread* currThread = NULL;
                for( unsigned int i = 0; i < p->threads.size(); i++ )
                {
		            if ((unsigned)p->threads[i]->get_tid() == debugEv.dwThreadId)
                    {
                        currThread = p->threads[i];
                        break;
                    }
                }
                assert( currThread != NULL );

                Address pendingTrampAddr = currThread->get_pending_tramp_addr();
                if( pendingTrampAddr != NULL )
                {
                    // reset the Eip to the pending instrumentation
                    CONTEXT ctxt;
                    ctxt.ContextFlags = CONTEXT_FULL;
                    if( !GetThreadContext( (HANDLE)currThread->get_handle(), &ctxt ) )
                    {
                        assert( false );
                    }
                    ctxt.Eip = pendingTrampAddr;
                    if( !SetThreadContext( (HANDLE)currThread->get_handle(), &ctxt ) )
                    {
                        assert( false );
                    }
                    currThread->set_pending_tramp_addr( NULL );
                }
            }
		    break;
	    }
	    p->status_ = running;
	    p->continueProc_();
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
			BPatch::bpatch->rDevice->RemoteContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
					   DBG_EXCEPTION_NOT_HANDLED);
#else
			ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
					   DBG_EXCEPTION_NOT_HANDLED);
#endif
	    return 0;

	case EXCEPTION_ACCESS_VIOLATION:
	    printf("Access violation exception, %d, addr = %08x\n", 
		   debugEv.u.Exception.ExceptionRecord.ExceptionFlags,
		   debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
	    dumpMem(p, debugEv.u.Exception.ExceptionRecord.ExceptionAddress, 32);

        {
#ifndef mips_unknown_ce2_11 //ccw 6 feb 2001 : 29 mar 2001
            vector<Address> pcs = p->walkStack( false );
            for( unsigned i = 0; i < pcs.size(); i++ )
            {
                function_base* f = p->findFunctionIn( pcs[i] );
                const char* szFuncName = (f != NULL) ? f->prettyName().string_of() : "<unknown>";
                fprintf( stderr, "%08x: %s\n", pcs[i], szFuncName );
            }
#endif
        }
	    //	ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
	    //			   DBG_EXCEPTION_NOT_HANDLED);
	    //	break;
	default:
	    //printf("exeption %x, addr %x\n", exnCode, 
	    //   debugEv.u.Exception.ExceptionRecord.ExceptionAddress);
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
			BPatch::bpatch->rDevice->RemoteContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
					   DBG_EXCEPTION_NOT_HANDLED);
#else
			ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
					   DBG_EXCEPTION_NOT_HANDLED);
#endif
	    return 0;
	}
    } break;

    case CREATE_THREAD_DEBUG_EVENT: {
	//printf("create thread, tid = %d\n", debugEv.dwThreadId);
	assert(p->threads.size() > 0); // main thread should be already defined
	pdThread *t = new pdThread(p, debugEv.dwThreadId,
				   debugEv.u.CreateThread.hThread);
	p->threads.push_back(t);
    } break;

    case CREATE_PROCESS_DEBUG_EVENT: {
	//CREATE_PROCESS_DEBUG_INFO info = debugEv.u.CreateProcessInfo;
	//printf("CREATE_PROCESS event: %d\n", debugEv.dwProcessId);
	p = findProcess(debugEv.dwProcessId);
	if (p) {
	    //fprintf(stderr,"create process: base = %x\n", info.lpBaseOfImage);
	    if (p->threads.size() == 0) {
		// define the main thread
		p->threads.push_back(new pdThread(p));
	    }
	    p->threads[0]->update_handle(debugEv.dwThreadId, 
					 debugEv.u.CreateProcessInfo.hThread);
	    
	}
    } break;

    case EXIT_THREAD_DEBUG_EVENT: {
	//printf("exit thread, tid = %d\n", debugEv.dwThreadId);
	unsigned nThreads = p->threads.size();
	// start from one to skip main thread
	for (unsigned u = 1; u < nThreads; u++) {
	    if ((unsigned)p->threads[u]->get_tid() == debugEv.dwThreadId) {
		delete p->threads[u];
		p->threads[u] = p->threads[nThreads-1];
		p->threads.resize(nThreads-1);
		break;
	    }
	}
    } break;

    case EXIT_PROCESS_DEBUG_EVENT:
	p = findProcess(debugEv.dwProcessId);
	if (p) {
            char errorLine[1024];
            sprintf(errorLine, "Process %d has terminated with code 0x%x\n", 
                p->getPid(), debugEv.u.ExitProcess.dwExitCode);
            statusLine(errorLine);
            logLine(errorLine);
	    handleProcessExit(p, debugEv.u.ExitProcess.dwExitCode);
	}
#ifdef BPATCH_LIBRARY
	*status = SIGEM_EXITED;
#else
	break;
#endif
	break; //ccw 23 july 2001
    case LOAD_DLL_DEBUG_EVENT: {
	//printf("load dll: hFile=%x, base=%x, debugOff=%x, debugSz=%d lpname=%x, %d\n",
	//       debugEv.u.LoadDll.hFile, debugEv.u.LoadDll.lpBaseOfDll,
	//       debugEv.u.LoadDll.dwDebugInfoFileOffset,
	//       debugEv.u.LoadDll.nDebugInfoSize,
	//       debugEv.u.LoadDll.lpImageName,
	//       /*debugEv.u.LoadDll.fUnicode*/
	//       GetFileSize(debugEv.u.LoadDll.hFile,NULL));

	// We are currently not handling dynamic libraries loaded after the
	// application starts.
	if (p->isBootstrappedYet())
	  break;

	// set the proc handle for the process that is loading the library
	// This is need when we use the imagehelp library to get the
	// symbols for the dll.
	// kludgeProcHandle = (HANDLE)p->getProcFileDescriptor();

        void *addr = debugEv.u.LoadDll.lpImageName;
        int ptr = 0;
	char nameBuffer[MAX_PATH];
	if (addr) {
	    /***
	      addr is a pointer to the image name (in the application address space)
	      There is a problem here, because this pointer can be null and it is
	      null in some cases (in particular, if we attach to a running process
	      it is always null). This means that it can be difficult to find
	      the name of an image. 

	      Currently, if we can't find the name, we just ignore the
	      image.  However, we need to at least be able to parse
	      kernel32.dll, because this library contains some very
	      important functions, including LoadLibrary, which we
	      need to call to load libdyninstRT.dll. We use the following
	      kludge to find if a library is kernel32.dll: the function
	      klugde_isKernel32Dll open a handle to kernel32.dll. We then
	      find if the handle we get from the LOAD_DLL_DEBUG_EVENT
	      is for the same file.
	    ***/

	    p->readDataSpace_(addr, 4, &ptr);
	    //printf("ptr = %x\n", ptr);
	    string kernel32Path;
		// ****************
		//ccw 10 aug 2000 : what is going on here???? ***************
		//at this point ptr contains the first 4 bytes of the name of the
		//DLL!?!?!?! IS DEBUGEVENT LOADDLL diff on nt/ce? MAYBE
		// ****************

	    if (ptr) {
#ifdef mips_unknown_ce2_11//ccw 11 aug 2000 : 29 mar 2001
				ptr = (unsigned int) addr;
#endif

		if (debugEv.u.LoadDll.fUnicode) {
		    unsigned short wbuffer[128];
		    p->readDataSpace_((void *)ptr, 128*sizeof(short), wbuffer);
		    WideCharToMultiByte(CP_ACP, 0, wbuffer, 128,
					nameBuffer, 128, NULL, NULL);
		} else {
		    p->readDataSpace_((void *)ptr, 128, nameBuffer);
		}
		//printf("Dll name: %s, base = %x\n", nameBuffer, debugEv.u.LoadDll.lpBaseOfDll);
	    }
	    else if (kludge_isKernel32Dll(debugEv.u.LoadDll.hFile, kernel32Path)) {
	        assert(kernel32Path.length() > 0);
	        assert(kernel32Path.length() < sizeof(nameBuffer));
		strcpy(nameBuffer, kernel32Path.string_of());
	    } else
	        break;
	}

	// try to load symbols for the DLL
        if (!SymLoadModule((HANDLE)p->getProcFileDescriptor(),
			   debugEv.u.LoadDll.hFile,
			   NULL, NULL, 0, 0)) {

		char msgText[1024];

		sprintf( msgText, "SymLoadModule failed for %s: 0x%x\n",
			nameBuffer, GetLastError() );

		logLine(msgText);
	}

	// discover structure of new DLL, and incorporate into our
	// list of known DLLs
	if (addr) {

	    shared_object *so = 
#ifdef mips_unknown_ce2_11 //ccw 29 mar 2001
		new shared_object(string(nameBuffer), (unsigned long) debugEv.u.LoadDll.lpBaseOfDll,
		false, true, true, 0);
#else
		new shared_object(string(nameBuffer), 0,false, true, true, 0);
#endif

	    assert(p->dyn);
	    p->dyn->sharedObjects.push_back(so);
	    if (!p->shared_objects) {
	      p->shared_objects = new vector<shared_object *>;
	    }
	    (*(p->shared_objects)).push_back(so);
#ifndef BPATCH_LIBRARY
	    tp->resourceBatchMode(true);
#endif 
	    p->addASharedObject(*so);
#ifndef BPATCH_LIBRARY
	    tp->resourceBatchMode(false);
#endif 
	    p->setDynamicLinking();
	  }

		//ccw 31 jan 2001 : 29 mar 2001
		//here i need to check to see if this is the first breakpt since
		//creating the process and if it is, do whatever should happen when 
		//i get the second NT breakpoint.
		//This is a HACK that fixes the bug that WinCE does NOT send a
		//debug message before starting main, you get one when you create
		//the process (before loading and dlls and any initializations take place)
		//and that is it.  so i grab one after coredll.dll is loaded
		//and load the libdyninstAPI_RT.dll

#ifdef mips_unknown_ce2_11 //ccw 31 jan 2001 : 29 mar 2001
		if(!strcmp(nameBuffer,"coredll.dll")){ 
			//make sure we have loaded coredll.dll --this contains LoadLibrary

			continueType = DBG_EXCEPTION_NOT_HANDLED;//ccw 25 oct 2000
			if (!p->reachedFirstBreak) {

				if (!p->hasLoadedDyninstLib && !p->isLoadingDyninstLib) {
					//DebugBreak();//ccw 30 jan 2001
					continueType = DBG_CONTINUE;//ccw 25 oct 2000
					//DebugBreak();
					Address addr = loadDyninstDll(p, p->savedData);
					p->savedRegs = p->getRegisters();
					p->changePC(addr);
					//p->LoadDyninstTrapAddr = addr + 0xd;
					p->isLoadingDyninstLib = true;
					break;
				}
			}
		}
#endif

      }
    break;
    case UNLOAD_DLL_DEBUG_EVENT:
	// TODO
	//printf("unload dll\n");
	break;
    case OUTPUT_DEBUG_STRING_EVENT:
#ifdef mips_unknown_ce2_11 //ccw 29 mar 2001
	//printf("output debug string\n");
		char dbgStr[512];//ccw 3 oct 2000
		memset(dbgStr,'\0', 512); //ccw 3 oct 2000
		p->readDataSpace_(debugEv.u.DebugString.lpDebugStringData, //ccw 3 oct 2000
			(debugEv.u.DebugString.nDebugStringLength<512) ? debugEv.u.DebugString.nDebugStringLength:511 , dbgStr );
		wprintf(L"output debug string: %s\n",dbgStr); //ccw 3 oct 2000
		//DebugBreak();
#endif
	break;
    case RIP_EVENT:
	//printf("rip event\n");
	break;
    default:
	;
	//printf("unknown debug event\n");
	
    }
    
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			    continueType)) {
#else
    if (!ContinueDebugEvent(debugEv.dwProcessId, debugEv.dwThreadId, 
			    DBG_CONTINUE)) {
#endif
	printf("ContinueDebugEvent failed\n");
	printSysError(GetLastError());
    }

#ifdef BPATCH_LIBRARY
    if (*status)
	return debugEv.dwProcessId;
    else
	return 0;
#else
    return 0;
#endif
}


#endif //ccw 21 july 2001

// already setup on this FD.
// disconnect from controlling terminal 
void OS::osDisconnect(void) {
#ifdef notdef
  int ttyfd = open ("/dev/tty", O_RDONLY);
  ioctl (ttyfd, TIOCNOTTY, NULL); 
  P_close (ttyfd);
#endif
}


bool process::attach() {
    if (createdViaAttach) {
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
	if (!BPatch::bpatch->rDevice->RemoteDebugActiveProcess(getPid())) {
#else
	if (!DebugActiveProcess(getPid())) {
#endif
	    //printf("Error: DebugActiveProcess failed\n");
	    return false;
	}
    }
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    proc_fd = (int)BPatch::bpatch->rDevice->RemoteOpenProcess(PROCESS_ALL_ACCESS, false, getPid());
#else
    proc_fd = (int)OpenProcess(PROCESS_ALL_ACCESS, false, getPid());

#endif

    if (proc_fd == NULL) {
	//printf("Error: OpenProcess failed\n");
	assert(0);
    }

    void initSymbols(HANDLE procH, const string file, const string dir);
    initSymbols((HANDLE)proc_fd, symbols->file(), "");
    if (createdViaAttach) {
	//
	//void initSymbols(HANDLE procH, const string file, const string dir);
	//initSymbols((HANDLE)proc_fd, symbols->file(), "");
    }
    return true;
}

/* continue a process that is stopped */
bool process::continueProc_() {
    if (hasNewPC) {
      changePC(currentPC_);
      hasNewPC = false;
    }
    for (unsigned u = 0; u < threads.size(); u++) {
	unsigned count;
#ifdef mips_unknown_ce2_11 //ccw 10 feb 2001 : 29 mar 2001
		count = BPatch::bpatch->rDevice->RemoteResumeThread((HANDLE)threads[u]->get_handle());
#else
		count = ResumeThread((HANDLE)threads[u]->get_handle());
#endif

	if (count == 0xFFFFFFFF) {
	    printSysError(GetLastError());
	    return false;
	}
    }
    return true;
}


#ifdef BPATCH_LIBRARY
/*
   terminate execution of a process
 */
bool process::terminateProc_()
{
    OS::osKill(pid);
    handleProcessExit(this, -1);
    return true;
}
#endif


/*
   pause a process that is running
*/
bool process::pause_() {
    for (unsigned u = 0; u < threads.size(); u++) {
	unsigned count ;
#ifdef mips_unknown_ce2_11 //ccw 10 feb 2001 : 29 mar 2001
	count = BPatch::bpatch->rDevice->RemoteSuspendThread((HANDLE)threads[u]->get_handle());
#else
	count = SuspendThread((HANDLE)threads[u]->get_handle());
#endif

	if (count == 0xFFFFFFFF) {
	    // printf("pause_: %d\n", threads[u]->get_tid());
	    // printSysError(GetLastError());
	    return false;
	} 
    }
    return true;
}

/*
   close the file descriptor for the file associated with a process
*/
bool process::detach_() {
    return false;
}


#ifdef BPATCH_LIBRARY
/*
   detach from thr process, continuing its execution if the parameter "cont"
   is true.
 */
bool process::API_detach_(const bool cont)
{
    // XXX Not yet implemented
    assert(false);
    return false;
}
#endif


bool process::dumpCore_(const string) {
    return false;
}

bool process::writeTextWord_(caddr_t inTraced, int data) {
    return writeDataSpace_(inTraced, sizeof(int), (caddr_t) &data);
}

bool process::writeTextSpace_(void *inTraced, u_int amount, const void *inSelf) {
    return writeDataSpace_(inTraced, amount, inSelf);
}

bool process::flushInstructionCache_(void *baseAddr, size_t size){ //ccw 25 june 2001
	return FlushInstructionCache((HANDLE)proc_fd, baseAddr, size);
}

#ifdef BPATCH_SET_MUTATIONS_ACTIVE
bool process::readTextSpace_(void *inTraced, u_int amount, const void *inSelf) {
  return readDataSpace_(inTraced, amount, (void *)inSelf);
}
#endif

bool process::writeDataSpace_(void *inTraced, u_int amount, const void *inSelf) {
    DWORD nbytes;

    //printf("write %d bytes, %x\n", amount, inTraced);
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    bool res = BPatch::bpatch->rDevice->RemoteWriteProcessMemory((HANDLE)proc_fd, (LPVOID)inTraced, 
				  (LPVOID)inSelf, (DWORD)amount, &nbytes);
#else
    bool res = WriteProcessMemory((HANDLE)proc_fd, (LPVOID)inTraced, 
				  (LPVOID)inSelf, (DWORD)amount, &nbytes);
#endif
    return res && (nbytes == amount);
}


bool process::readDataSpace_(const void *inTraced, u_int amount, void *inSelf) {
    DWORD nbytes;
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    bool res = BPatch::bpatch->rDevice->RemoteReadProcessMemory((HANDLE)proc_fd, (LPVOID)inTraced, 
				 (LPVOID)inSelf, (DWORD)amount, &nbytes);
#else
    bool res = ReadProcessMemory((HANDLE)proc_fd, (LPVOID)inTraced, 
				 (LPVOID)inSelf, (DWORD)amount, &nbytes);
#endif
    return res && (nbytes == amount);
}


bool process::loopUntilStopped() { assert(0); return false; }

float OS::compute_rusage_cpu() { return 0.0; }
float OS::compute_rusage_sys() { return 0.0; }
float OS::compute_rusage_min() { return 0.0; }
float OS::compute_rusage_maj() { return 0.0; }
float OS::compute_rusage_swap() { return 0.0; }
float OS::compute_rusage_io_in() { return 0.0; }
float OS::compute_rusage_io_out() { return 0.0; }
float OS::compute_rusage_msg_send() { return 0.0; }
float OS::compute_rusage_msg_recv() { return 0.0; }
float OS::compute_rusage_sigs() { return 0.0; }
float OS::compute_rusage_vol_cs() { return 0.0; }
float OS::compute_rusage_inv_cs() { return 0.0; }


int getNumberOfCPUs() {
    SYSTEM_INFO info;
    GetSystemInfo(&info);
    return info.dwNumberOfProcessors;
}  


void Frame::getActiveFrame(process *p)
{
   w32CONTEXT cont; //ccw 27 july 2000 : 29 mar 2001

    // we must set ContextFlags to indicate the registers we want returned,
    // in this case, the control registers.
    // The values for ContextFlags are defined in winnt.h
    cont.ContextFlags = CONTEXT_CONTROL;
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (BPatch::bpatch->rDevice->RemoteGetThreadContext((HANDLE)p->threads[0]->get_handle(), &cont)) {
#else
    if (GetThreadContext((HANDLE)p->threads[0]->get_handle(), &cont)) {

#endif

#ifdef i386_unknown_nt4_0 //ccw 27 july 2000 : 29 mar 2001

        fp_ = cont.Ebp;
	pc_ = cont.Eip;
#elif mips_unknown_ce2_11
	pc_ = cont.Fir;
	sp_ = cont.IntSp;
	fp_ = cont.IntS8;
#endif

	return;
    }
    printSysError(GetLastError());
}

#if defined(i386_unknown_nt4_0) //ccw 29 mar 2001
 

Frame Frame::getCallerFrameNormal(process *p) const
{
    //
    // for the x86, the frame-pointer (EBP) points to the previous frame-pointer,
    // and the saved return address is in EBP-4.
    //
    struct {
	Address fp;
	Address rtn;
    } addrs;

    if (p->readDataSpace((caddr_t)(fp_), sizeof(int)*2,
			 (caddr_t)&addrs, true))
    {
        Frame ret;
        ret.fp_ = addrs.fp;
	ret.pc_ = addrs.rtn;
	return ret;
    }

    return Frame(); // zero frame
}

#elif mips_unknown_ce2_11 //ccw 6 feb 2001

Frame Frame::getCallerFrameNormal(process *p) const
{
//	DebugBreak();//ccw 6 feb 2001

	Frame ret;
	Address prevPC;

	//find which function we are in to determine the frame size

	// (i.e. $pc in native/basetramp/minitramp code)
	instPoint     *ip = NULL;
	trampTemplate *bt = NULL;
	instInstance  *mt = NULL;
	pd_Function *callee = findAddressInFuncsAndTramps(p, pc_, ip, bt, mt);
	// non-NULL "ip" means that $pc is in instrumentation
	// non-NULL "bt" means that $pc is in basetramp
	// non-NULL "mt" means that $pc is in minitramp
	if (ip) assert(bt || mt);

	ret.sp_ = sp_ - callee->frame_size;

	Address tmpSp = sp_ + 20;
	p->readDataSpace((caddr_t)(tmpSp), sizeof(int),
			 &prevPC, true);

	ret.pc_ = prevPC;
	  

	return ret;
}
#endif

#ifdef mips_unknown_ce2_11 //ccw 10 aug 2000 : 29 mar 2001
void *process::getRegisters(unsigned int thrHandle) {
    w32CONTEXT *cont = new w32CONTEXT; //ccw 27 july 2000
    if (!cont)
		return NULL;
    // we must set ContextFlags to indicate the registers we want returned,
    // in this case, the control registers.
    // The values for ContextFlags are defined in winnt.h
    cont->ContextFlags = w32CONTEXT_FULL;//ccw 27 july 2000
    if (!BPatch::bpatch->rDevice->RemoteGetThreadContext((void*) thrHandle, cont)) { //ccw 10 aug 2000
		delete cont;
		return NULL;
    }
    return (void *)cont;
}
#endif

void *process::getRegisters() {
    w32CONTEXT *cont = new w32CONTEXT; //ccw 27 july 2000 : 29 mar 2001
    if (!cont)
	return NULL;
    // we must set ContextFlags to indicate the registers we want returned,
    // in this case, the control registers.
    // The values for ContextFlags are defined in winnt.h
    cont->ContextFlags = w32CONTEXT_FULL;//ccw 27 july 2000 : 29 mar 2001
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteGetThreadContext((HANDLE)threads[0]->get_handle(), cont)) {
#else
    if (!GetThreadContext((HANDLE)threads[0]->get_handle(), cont)) {
#endif

	delete cont;
	return NULL;
    }
    return (void *)cont;
}

bool process::changePC(Address addr, const void *savedRegs) {
    assert(status_ == stopped);
    w32CONTEXT cont = *(w32CONTEXT *)savedRegs; //ccw 27 july 2000 : 29 mar 2001
//	DebugBreak();
#ifdef i386_unknown_nt4_0 //ccw 27 july 2000 : 29 mar 2001
    cont.Eip = addr;
#elif  mips_unknown_ce2_11 
	cont.Fir = addr;
#endif

#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteSetThreadContext((HANDLE)threads[0]->get_handle(), &cont)) {
#else
    if (!SetThreadContext((HANDLE)threads[0]->get_handle(), &cont)) {
#endif

	printf("SethreadContext failed\n");
	return false;
    }
    return true;
}

bool process::changePC(Address addr) {
    assert(status_ == stopped || status_ == neonatal);

     w32CONTEXT cont;//ccw 27 july 2000

    cont.ContextFlags = w32CONTEXT_FULL;//ccw 27 july 2000 : 29 mar 2001
//	DebugBreak();
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteGetThreadContext((HANDLE)threads[0]->get_handle(), &cont)) {
#else
    if (!GetThreadContext((HANDLE)threads[0]->get_handle(), &cont)) {
#endif
		printf("GetThreadContext failed\n");
		return false;
    }
#ifdef i386_unknown_nt4_0 //ccw 27 july 2000 : 29 mar 2001
    cont.Eip = addr;
#elif  mips_unknown_ce2_11 
	cont.Fir = addr;
#endif
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteSetThreadContext((HANDLE)threads[0]->get_handle(), &cont)) {
#else
    if (!SetThreadContext((HANDLE)threads[0]->get_handle(), &cont)) {
#endif
	return false;
    }
    return true;
}

bool process::restoreRegisters(void *buffer) {
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteSetThreadContext((HANDLE)threads[0]->get_handle(),
	 (w32CONTEXT *)buffer)) {//ccw 27 july 2000
#else
    if (!SetThreadContext((HANDLE)threads[0]->get_handle(), (w32CONTEXT *)buffer)) {//ccw 27 july 2000
#endif
	//printf("SetThreadContext failed\n");
	return false;
    }
    return true;
}

bool process::isRunning_() const {
    // TODO
    printf("process::isRunning_() returning true\n");
    return true;
}


string process::tryToFindExecutable(const string& iprogpath, int pid) {
  return iprogpath;
}

bool process::set_breakpoint_for_syscall_completion() {
   /* Can assume: (1) process is paused and (2) in a system call.
      We want to set a TRAP for the syscall exit, and do the
      inferiorRPC at that time.  We'll use /proc PIOCSEXIT.
      Returns true iff breakpoint was successfully set. */

    // This is never called on Windows NT
    assert(false);
    return false;
}

void process::clear_breakpoint_for_syscall_completion() { return; }

Address process::read_inferiorRPC_result_register(Register reg) { //ccw 17 july 2001
   w32CONTEXT *cont = new w32CONTEXT;//ccw 27 july 2000 : 29 mar 2001
    if (!cont)
	return NULL;
    // we must set ContextFlags to indicate the registers we want returned,
    // in this case, the control registers.
    // The values for ContextFlags are defined in winnt.h
    cont->ContextFlags = w32CONTEXT_FULL;//ccw 27 july 2000
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    if (!BPatch::bpatch->rDevice->RemoteGetThreadContext((HANDLE)threads[0]->get_handle(), cont)) {
#else
    if (!GetThreadContext((HANDLE)threads[0]->get_handle(), cont)) {
#endif
	delete cont;
	return NULL;
    }
#ifdef i386_unknown_nt4_0 //ccw 27 july 2000 : 29 mar 2001
    return cont->Eax;
#elif mips_unknown_ce2_11 
	return (&cont->IntZero)[reg]; 
	//the registers in the MIPS context returned by
	//winCE return general registers 0..31 in the struct
	//starting at IntZero and going to IntRa, each being a DWORD
#endif
}

bool process::executingSystemCall() {
   // TODO
   return false;
}

// TODO
bool process::needToAddALeafFrame(Frame , Address &) {
  return false;
}


void initSymbols(HANDLE procH, const string file, const string dir) {
  string searchPath;
  char sysRootDir[MAX_PATH+1];
  if (GetSystemDirectory(sysRootDir, MAX_PATH) == 0)
     assert(0);
  string sysSymsDir = string(sysRootDir) + "\\..\\symbols";
  if (dir.length())
    searchPath = dir + ";";
  searchPath = searchPath + sysSymsDir + ";" + sysSymsDir + "\\dll";
  if (!SymInitialize(procH, (char *)searchPath.string_of(), 0)) {
    fprintf(stderr,"SymInitialize failed, %x\n", GetLastError()); fflush(stderr);
    return;
  }
  if (!SymLoadModule(procH, NULL, (char *)file.string_of(), NULL, 0, 0)) {
    printf("SymLoadModule failed for \"%s\", %x\n",
	    file.string_of(), GetLastError());
    return;
  }
}





/*****************************************************************************
 * forkNewProcess: starts a new process, setting up trace and io links between
 *                the new process and the daemon
 * Returns true if succesfull.
 * 
 * Arguments:
 *   file: file to execute
 *   dir: working directory for the new process
 *   argv: arguments to new process
 *   envp: environment **** not in use
 *   inputFile: where to redirect standard input
 *   outputFile: where to redirect standard output
 *   traceLink: handle or file descriptor of trace link (read only)
 //removed all ioLink related code for output redirection
 *   ioLink: handle or file descriptor of io link (read only)
 *   pid: process id of new process
 *   tid: thread id for main thread (needed by WindowsNT)
 *   procHandle: handle for new process (needed by WindowsNT)
 *   thrHandle: handle for main thread (needed by WindowsNT)
 ****************************************************************************/
bool forkNewProcess(string &file, string dir, vector<string> argv, 
		    vector<string>envp, string inputFile, string outputFile,
		    int &traceLink,  
		    int &pid, int &tid, 
		    int &procHandle, int &thrHandle, int /* stdin_fd */, 
		    int stdout_fd, int /* stderr_fd */) {
#ifdef mips_unknown_ce2_11 //ccw 8 aug 2000 : 29 mar 2001
	WCHAR appName[256];
	WCHAR dirName[256];
	WCHAR commLine[256];
#endif

#ifndef BPATCH_LIBRARY
#ifdef notdef_Pipes     // isn't this all obsolete???
    HANDLE rTracePipe;
    HANDLE wTracePipe;
    // security attributes to make handles inherited
    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.bInheritHandle = true;
    sa.lpSecurityDescriptor = NULL;
    
    // create trace pipe
    if (!CreatePipe(&rTracePipe, &wTracePipe, &sa, 0)) {
	string msg = string("Unable to create trace pipe for program '") 
	    + File + string("': ") + string(sys_errlist[errno]);
	showErrorCallback(68, msg);
	return(NULL);
    }
    
    /* removed for output redirection
    HANDLE rIoPipe;
    HANDLE wIoPipe;
    // create IO pipe
    // ioPipe is used to redirect the child's stdout & stderr to a pipe which
    // is in turn read by the parent via the process->ioLink socket.
    if (!CreatePipe(&rIoPipe, &wIoPipe, &sa, 0)) {
	string msg = string("Unable to create IO pipe for program '") + File +
	    string("': ") + string(sys_errlist[errno]);
	showErrorCallback(68, msg);
	return(NULL);
    }
    SetEnvironmentVariable("PARADYN_IO_PIPE",
			   string((unsigned)wIoPipe).string_of());
    */

    printf("tracepipe = %d\n", (unsigned)wTracePipe);
    // enter trace and IO pipes in child's environment
    SetEnvironmentVariable("PARADYN_TRACE_PIPE", 
			   string((unsigned)wTracePipe).string_of());
#endif
    //  extern int traceSocket;
    //  SetEnvironmentVariable("PARADYND_TRACE_SOCKET", string((unsigned)traceSocket).string_of());
#endif /* BPATCH_LIBRARY */
    
    // create the child process
    
    string args;
    for (unsigned ai=0; ai<argv.size(); ai++) {
	args += argv[ai];
	args += " ";
    }
    
    STARTUPINFO stinfo;
    memset(&stinfo, 0, sizeof(STARTUPINFO));
    stinfo.cb = sizeof(STARTUPINFO);

    /*to do: output redirection
    //stinfo.hStdOutput = (HANDLE)ioLink;
    stinfo.hStdOutput = (HANDLE)stdout_fd;
    stinfo.hStdInput = GetStdHandle(STD_INPUT_HANDLE);
    stinfo.hStdError = GetStdHandle(STD_ERROR_HANDLE);
    stinfo.dwFlags |= STARTF_USESTDHANDLES;
    */
    PROCESS_INFORMATION procInfo;
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
	kludgeWCHAR(file.string_of(),appName);
	kludgeWCHAR(args.string_of(),commLine);
	if(dir == ""){
		dirName[0] = (unsigned short) 0;
	}else{
		kludgeWCHAR(dir.string_of(),dirName);
	}

	/* ccw 8 aug 2000  : 29 mar 2001
	 WARNING:  
		If compiler optimization are turned on (-Ox) for remoteDevice.C the following
		line of code bugs. An extra parameter is passed to the function before 
		appName and consequently each argument is shifted one to the right, 
		completely destroying the code.
		
		DO NOT TURN ON -Ox for remoteDevice.C!

		if need be, pdwinnt.C can be compiled with -Ox
		less optimization may be safe, it has not yet been tried. if someone is bored
		they should try different optimization flags to see if any work correctly here.
		Until then the mips/ce version of dyninst will be compiled w/o ANY optimizations!
	*/
	//DebugBreak();
    if (BPatch::bpatch->rDevice->RemoteCreateProcess(appName , commLine,
		      NULL, NULL, false,
		      (DWORD) DEBUG_PROCESS  /* | CREATE_NEW_CONSOLE /* | CREATE_SUSPENDED */,
		      NULL, dirName, 
		      &stinfo, &procInfo)) {

#else

    if (CreateProcess(file.string_of(), (char *)args.string_of(), 
		      NULL, NULL, TRUE,
		      DEBUG_PROCESS /* | CREATE_NEW_CONSOLE /* | CREATE_SUSPENDED */,
		      NULL, dir == "" ? NULL : dir.string_of(), 
		      &stinfo, &procInfo)) {
#endif

	procHandle = (Word)procInfo.hProcess;
	thrHandle = (Word)procInfo.hThread;
	pid = (Word)procInfo.dwProcessId;
	tid = (Word)procInfo.dwThreadId;
#ifdef notdef_Pipes
        traceLink = (Word)rTracePipe;
	CloseHandle(wTracePipe);
	/ *removed for output redirection
        //ioLink = (Word)rIoPipe;
        CloseHandle(wIoPipe);
        */
        //initSymbols((HANDLE)procHandle, file, dir);
#else
        traceLink = -1;
	/* removed for output redirection
        ioLink = -1;
	*/
#endif
	return true;
    }
   
#ifndef BPATCH_LIBRARY
#ifdef notdef_Pipes
    CloseHandle(rTracePipe);
    CloseHandle(wTracePipe);
    / *removed for output redirection
    CloseHandle(rIoPipe);
    CloseHandle(wIoPipe);
    */
#endif
#endif /* BPATCH_LIBRARY */

    // Output failure message
    LPVOID lpMsgBuf;

    if (FormatMessage( 
	    FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
	    NULL,
	    GetLastError(),
	    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
	    (LPTSTR) &lpMsgBuf,
	    0,
	    NULL 
	) > 0) {
	char *errorLine = (char *)malloc(strlen((char *)lpMsgBuf) +
					 file.length() + 64);
	if (errorLine != NULL) {
	    sprintf(errorLine, "Unable to start %s: %s\n", file.string_of(),
		    (char *)lpMsgBuf);
	    logLine(errorLine);
	    showErrorCallback(68, (const char *) errorLine);

	    free(errorLine);
	}

	// Free the buffer returned by FormatMsg
	LocalFree(lpMsgBuf);
    } else {
	char errorLine[512];
	sprintf(errorLine, "Unable to start %s: unknown error\n",
		file.string_of());
	logLine(errorLine);
	showErrorCallback(68, (const char *) errorLine);
    }

    return false;
}

char *cplus_demangle(char *c, int) { 
    char buf[1000];
    if (c[0]=='_') {
        // VC++ 5.0 seems to decorate C symbols differently to C++ symbols
        // and the UnDecorateSymbolName() function provided by imagehlp.lib
        // doesn't manage (or want) to undecorate them, so it has to be done
        // manually, removing a leading underscore from functions & variables
        // and the trailing "$stuff" from variables (actually "$Sstuff")
        unsigned i;
        for (i=1; i<sizeof(buf) && c[i]!='$' && c[i]!='\0'; i++)
           { buf[i-1]=c[i]; }
        if (i==1) return 0; // avoid null names which seem to annoy Paradyn
        buf[i-1]='\0';
        return strdup(buf);
      }
    else if (UnDecorateSymbolName(c, buf, 1000, UNDNAME_NAME_ONLY)) {
	//printf("Undecorate: %s = %s\n", c, buf);

        // many symbols have a name like foo@4, we must remove the @4
        // just searching for an @ is not enough,
        // as it may occur on other positions. We search for the last one
        // and check that it is followed only by digits.
        char *p = strrchr(buf, '@');
	if (p) {
	  char *q = p+1;
	  strtoul(p+1, &q, 10);
	  if (q > p+1 && *q == '\0') {
	    *p = '\0';
	  }
	}
	return strdup(buf);
    }
    return 0;
}

bool OS::osKill(int pid) {
    bool res;
#ifdef mips_unknown_ce2_11 //ccw 28 july 2000 : 29 mar 2001
    HANDLE h = BPatch::bpatch->rDevice->RemoteOpenProcess(PROCESS_ALL_ACCESS, false, pid);
#else
    HANDLE h = OpenProcess(PROCESS_ALL_ACCESS, false, pid);
#endif
    if (h == NULL) {
	return false;
    }
#ifdef mips_unknown_ce2_11 //ccw 31 aug 2000 : 29 mar 2001
    res = BPatch::bpatch->rDevice->RemoteTerminateProcess(h,0);
    BPatch::bpatch->rDevice->RemoteCloseHandle(h);
#else
    res = TerminateProcess(h,0);
    CloseHandle(h);
#endif
    return res;
}

#ifdef SHM_SAMPLING
rawTime64 process::getRawCpuTime_hw(int lwp_id) {
  lwp_id = 0;  // to turn off warning for now
  return 0;
}

rawTime64
FILETIME2rawTime64( FILETIME& ft )
{
    return (((rawTime64)(ft).dwHighDateTime<<32) | 
	    ((rawTime64)(ft).dwLowDateTime));
}

/* return unit: nsecs */
rawTime64 process::getRawCpuTime_sw(int lwp_id) {
  FILETIME kernelT, userT, creatT, exitT;
  rawTime64 now;

  if(GetProcessTimes( (HANDLE)getProcFileDescriptor(),
      &creatT, &exitT, &kernelT,&userT)==0) {
    abort();
    return 0;
  }

  // GetProcessTimes returns values in 100-ns units
  now = (FILETIME2rawTime64(userT)+FILETIME2rawTime64(kernelT));

  // time shouldn't go backwards, but we'd better handle it if it does
  if (now < previous) {
     logLine("********* time going backwards in paradynd **********\n");
     now = previous;
  }
  else {
     previous = now;
  }

  return now;
}
#endif // SHM_SAMPLING

#ifdef mips_unknown_ce2_11 //ccw 2 aug 2000 : 29 mar 2001

#include "inst-mips.h"
//this comes from irix.C
bool process::heapIsOk(const vector<sym_data>&findUs)
{ 
	
  //ccw 12 oct 2000 NEED TO FIND WinMain here or _WinMain then try to find main or _main
//	DebugBreak();


  if (!(mainFunction = findOneFunction("WinMain")) &&
	  !(mainFunction = findOneFunction("_WinMain"))) {
		fprintf(stderr, "process::heapIsOk(): failed to find \"WinMain\"\n");

		if (!(mainFunction = findOneFunction("main")) &&
		!(mainFunction = findOneFunction("_main"))) {

			//if (!(mainFunction = findOneFunction("CFrameWnd::OnCreate")) &&
			//	!(mainFunction = findOneFunction("OnCreate"))) {

				//the following is for MFC apps
				//what is wrong with M$? all these different names for main()
				if (!(mainFunction = findOneFunction("wWinMain")) &&
					!(mainFunction = findOneFunction("_wWinMain"))) {


					fprintf(stderr, "process::heapIsOk(): failed to find \"main\"\n");
					return false;
				}
			//}
	  }	
  }

  for (unsigned i = 0; i < findUs.size(); i++) {
    const string &name = findUs[i].name;
    cerr << "Looking up symbol " << name << endl;
    /*
    Address addr = lookup_fn(this, name);
    if (!addr && findUs[i].must_find) {
      fprintf(stderr, "process::heapIsOk(): failed to find \"%s\"\n", name.string_of());
      return false;
    }
    */
  }

  return true;
}
#endif


fileDescriptor *getExecFileDescriptor(string filename,
				     int &,
				     bool)
{
  fileDescriptor *desc = new fileDescriptor(filename);
  return desc;
}

#ifndef BPATCH_LIBRARY
void process::initCpuTimeMgrPlt() {
  cpuTimeMgr->installLevel(cpuTimeMgr_t::LEVEL_TWO, &process::yesAvail, 
			   timeUnit(fraction(100)), timeBase::bNone(), 
			   &process::getRawCpuTime_sw, "swCpuTimeFPtrInfo");
}
#endif
