/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */





/************************************************************************
 * queue.h: header file for a FIFO queue ADT.
************************************************************************/





#if !defined(_thread_src_queue_h_)
#define _thread_src_queue_h_





/************************************************************************
 * structures and types.
************************************************************************/

typedef struct thr_queue* Queue;
struct thr_queue {
    unsigned size;
    unsigned head;
    unsigned tail;
    void**   items;
};

typedef unsigned qscan_t;





/************************************************************************
 * macros.
************************************************************************/

#define NEWQ(q)       (q = q_init(0))
#define FREEQ(q)      q_free(q)
#define CLEARQ(q)     q_clear(q)
#define ISEMPTYQ(q)   q_isempty(q)
#define COPYQ(d,s)    q_copy(d, s)
#define ENQUEUE(q, d) q_enqueue(q, (void *) d)
#define DEQUEUE(q)    q_dequeue(q)
#define PRINTQ(q)     q_print(q)
#define FIRST(q)      q_initscan(q)
#define EOSCAN(q, qs) q_isbadqs(q, qs)
#define NEXT(q, qs)   q_nextscan(q, qs)
#define GET(q, qs)    q_getitem(q, qs)
#define REMOVE(q, qs) q_delitem(q, qs)





/************************************************************************
 * function prototypes.
************************************************************************/

Queue    q_init(unsigned);
void     q_free(Queue);
void     q_clear(Queue);
void     q_copy(Queue, Queue);
void*    q_dequeue(Queue);
void     q_enqueue(Queue, void *);
bool     q_isempty(Queue);
void     q_print(Queue);

qscan_t  q_initscan(Queue);
qscan_t  q_nextscan(Queue, qscan_t);
bool     q_isbadqs(Queue, qscan_t);
void*    q_getitem(Queue, qscan_t);
void*    q_delitem(Queue, qscan_t);





#endif /* !defined(_thread_src_queue_h_) */
