/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */




/************************************************************************
 * threadP.h: private header file for libthread.
************************************************************************/





#if !defined(_thread_src_threadp_h_)
#define _thread_src_threadp_h_





/************************************************************************
 * header files.
************************************************************************/

#include <assert.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#if defined(THREADLIB_DEBUG)
#include <iostream.h>
#endif // defined(THREADLIB_DEBUG)

#if !defined(i386_unknown_nt4_0)
#include <sys/param.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <unistd.h>
#else
#include <windows.h>
#include <winsock2.h>
#include <io.h>
#endif /* !defined(i386_unknown_nt4_0) */

#ifdef rs6000_ibm_aix4_1
#include <sys/select.h> // for fd_set
#endif

#include "pdutil/h/pddesc.h"
#include "thread/h/thread.h"
#include "kludges.h"
#include "queue.h"
#include "waitset.h"
#include "thrmsg.h"
#include "itembase.h"
#include "thr.h"
#include "file.h"
#include "sock.h"
#include "sig.h"
#if defined(i386_unknown_nt4_0)
#include "wmsg.h"
#endif // defined(i386_unknown_nt4_0)
#include "thrtab.h"





/************************************************************************
 * constants.
************************************************************************/


#define THR_ENOMEM		(THR_ERR-1)
#define THR_EDEADLK		(THR_ERR-2)
#define THR_ESRCH		(THR_ERR-3)
#define THR_EBADKEY		(THR_ERR-4)
#define THR_EBADFD		(THR_ERR-5)
#define THR_EBADBF		(THR_ERR-6)
#define THR_EBADTAG		(THR_ERR-7)
#define THR_EBADSIG		(THR_ERR-8)
#define THR_ENOMSG		(THR_ERR-9)
#define	THR_EBADSOCK	(THR_ERR-10)
#define	THR_EPARAM		(THR_ERR-11)

#define PRINTF_BUFSIZE 8192




/************************************************************************
 * structures and types.
************************************************************************/

typedef void (*Destructor)(void *);





/************************************************************************
 * macros.
************************************************************************/

#define CINCR(v,s) (((v)+1)%(s))
#define CDECR(v,s) (((v)+(s)-1)%(s))

#define abs(x)     (((x)>0)?(x):(-(x)))

#if !defined(min)
#define min(x,y)   (((x)<(y))?(x):(y))
#endif /* !defined(min) */

#if !defined(max)
#define max(x,y)   (((x)>(y))?(x):(y))
#endif /* !defined(max) */

#define sig2ndx(x) ((x)-1)
#define ndx2sig(x) ((x)+1)

#define	is_unspec_tid(x)	(x == THR_TID_UNSPEC)
#define	is_unspec_tag(x)	(x == MSG_TAG_UNSPEC)

#define is_invalid_key(k) ((k) >= thr_curr_key)

#define is_sig_tag(t)	(is_unspec_tag(t)||(t==MSG_TAG_SIG))
#define	is_io_tag(t)	(is_unspec_tag(t)|| \
						(t==MSG_TAG_FILE)|| \
						(t==MSG_TAG_SOCKET))
#define	is_file_tag(t)	(is_unspec_tag(t)||(t==MSG_TAG_FILE))
#define	is_spl_tag(t)	(t>=0)
#define is_thr_tag(t)	(is_unspec_tag(t)||(t==MSG_TAG_THREAD)||(t>=MSG_TAG_USER))
#define	is_socket_tag(t)	(is_unspec_tag(t)||(t==MSG_TAG_SOCKET))

#define thr_err(x) (((thr_curr->id.errno_local)=(x)),(((x)==THR_OKAY)?(x):(THR_ERR)))

#if defined(ENABLE_TRACE)
#define DO_TRACE_0(x) thr_do_trace_0 x
#define DO_TRACE_1(x) thr_do_trace_1 x
#else /* defined(ENABLE_TRACE) */
#define DO_TRACE_0(x)
#define DO_TRACE_1(x)
#endif /* defined(ENABLE_TRACE) */

#if defined(ALWAYS_RESCHEDULE)
#define CHECK_RESCHEDULE()              \
    do {                                \
        ENQUEUE(thr_ready_q, thr_curr); \
        thr_reschedule();               \
    } while (0)
#else /* defined(ALWAYS_RESCHEDULE) */
#define CHECK_RESCHEDULE()
#endif /* defined(ALWAYS_RESCHEDULE) */

#if defined(ALWAYS_REALLY_RESCHEDULE)
#define CHECK_REALLY_RESCHEDULE()       \
    do {                                \
        ENQUEUE(thr_ready_q, thr_curr); \
        thr_reschedule();               \
    } while (0)
#else /* defined(ALWAYS_REALLY_RESCHEDULE) */
#define CHECK_REALLY_RESCHEDULE()
#endif /* defined(ALWAYS_REALLY_RESCHEDULE) */

#define RETURN(x)     \
do {                  \
    ret = thr_err(x); \
    goto finish;      \
} while (0)

#define VRETURN       \
do {                  \
    goto finish;      \
} while (0)

#if defined(ENABLE_PREEMPTION)

#define THR_SAVE_CONTEXT(env)    thr_save_context(env, 1)
#define THR_RESTORE_CONTEXT(env) thr_restore_context(env, 1)

#define NOT_VTALRM(signo)                                  \
do {                                                       \
    if (signo == SIGVTALRM) {                              \
        fprintf(stderr, "signal `%d' used by libthread\n", \
            signo);                                        \
        RETURN(THR_EBADSIG);                               \
    }                                                      \
} while (0)

#define DUMMY_ENTER()       \
do {                        \
    assert(thr_inlib > 0);  \
    thr_inlib = 0;          \
} while (0)

#define DUMMY_LEAVE()       \
do {                        \
    assert(thr_inlib == 0); \
} while (0)

#define THR_ENABLE_PREEMPTION() thr_enable_preemption()

#define THR_ENTER_LIB() (thr_inlib++)
#define THR_LEAVE_LIB() (thr_inlib--)

#define SAVE_INLIB()              \
do {                              \
    unsigned save_inlib;          \
                                  \
    do {                          \
        thr_inlib++;              \
        save_inlib = thr_inlib-1; \
    } while (0)

#define RESTORE_INLIB()           \
    thr_inlib = save_inlib;       \
} while (0)

#else /* defined(ENABLE_PREEMPTION) */


#if defined(i386_unknown_nt4_0)
#  define THR_SAVE_CONTEXT(env)		(0)
#  define THR_RESTORE_CONTEXT(fiber)		\
	do										\
	{										\
		if( fiber != GetCurrentFiber() )	\
		{									\
			SwitchToFiber(fiber);			\
		}									\
	} while (false)

#else // defined(i386_unknown_nt4_0)
#  if defined(i386_unknown_linux2_0)
  /* Under Linux glibc, setjmp(env) is only a macro to sigsetjmp(env,0),
   * so a function pointer is not possible.  A wrapper function also
   * doesn't help as setjmp restores only the stack pointer but not
   * the stack contents, and you get seg faults. -DAN 4/21/98
   */
#    define THR_SAVE_CONTEXT(env)    setjmp(env)
#  else
#    define THR_SAVE_CONTEXT(env)    Tcb::save_context(env)
#  endif
#  define THR_RESTORE_CONTEXT(env) Tcb::restore_context(env, 1)
#endif // defined(i386_unknown_nt4_0)

#define NOT_VTALRM(signo)

#define DUMMY_ENTER()
#define DUMMY_LEAVE()

#define THR_ENABLE_PREEMPTION()

#define THR_ENTER_LIB()
#define THR_LEAVE_LIB()

#define SAVE_INLIB()
#define RESTORE_INLIB()

#endif /* defined(ENABLE_PREEMPTION) */




/************************************************************************
 * globals.
************************************************************************/

extern unsigned       thr_trace_flag;
extern FILE*          thr_trace_fp;

#if defined(ENABLE_PREEMPTION)
extern unsigned       thr_inlib;
extern unsigned       thr_quantum;
#endif /* defined(ENABLE_PREEMPTION) */

extern unsigned       thr_init_done;
extern unsigned       thr_pid;

extern ThrTable*      thr_tab;

extern Queue          thr_ready_q;
extern Queue          thr_zombie_q;
extern Queue          thr_waiter_q;
extern Thread         thr_curr;

extern unsigned       thr_curr_key;
extern Destructor     thr_key_destructors[];

extern WaitSet        thr_all_waitset;
extern thread_t       thr_last;

extern const char*    thr_errlst[];
extern const unsigned thr_nerr;





/************************************************************************
 * inline utility functions
************************************************************************/

inline
bool
SpecifiesThread(thread_t tid, tag_t tag)
{
	return ((is_unspec_tid(tid) && is_thr_tag(tag)) || 
			(!is_unspec_tid(tid) && thr_tab->IsThread(tid)));
}


inline
bool
SpecifiesFile(thread_t tid, tag_t tag)
{
	return ((is_unspec_tid(tid) && is_file_tag(tag)) || 
			(!is_unspec_tid(tid) && thr_tab->IsFile(tid)));
}


inline
bool
SpecifiesSocket(thread_t tid, tag_t tag)
{
	return ((is_unspec_tid(tid) && is_socket_tag(tag)) || 
			(!is_unspec_tid(tid) && thr_tab->IsSocket(tid)));
}







/************************************************************************
 * function prototypes.
************************************************************************/

unsigned	thr_check_sig_msg(void);
bool		thr_is_ready_io(const WaitSet& ws, bool, thread_t, thread_t*, tag_t*);
void		thr_adjust_stack(Thread);
void		thr_check_stack(void);
void		thr_cleanup(void *);
void		thr_do_arch_magic(Thread);
void		thr_init(void);
void		thr_init_arch_magic(void);
void		thr_init_sig_map(unsigned);
void		thr_print_tcb(void);
void		thr_reschedule(void);
void		thr_service_io(void);
#if SIG_READY
void		thr_sig_enter_cs(void);
void		thr_sig_leave_cs(void);
#endif /* SIG_READY */

void		thr_do_trace_0(const char *, ...);
void		thr_do_trace_1(const char *, ...);

#if defined(ENABLE_PREEMPTION)
void		thr_enable_preemption(void);
#endif /* defined(ENABLE_PREEMPTION) */

void		thr_report_error(const char* fmt, ...);



#endif /* !defined(_thread_src_threadp_h_) */
